import sys
import subprocess
import os
import argparse
import lmql

import lmql.version as version_info

project_root = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

def cmd_serve_model():
    """emoji:🏄 Serve a 🤗 Transformers model via the LMQL inference API"""
    os.chdir(project_root)
    os.system("python -m lmql.model.serve " + " ".join(sys.argv[2:]))

def cmd_run():
    """
    emoji:🏃 run a LMQL script (e.g. "lmql run latest/hello.lmql")
    """
    import asyncio
    import time

    start = time.time()

    parser = argparse.ArgumentParser(description="Runs a LMQL program.")
    parser.add_argument("lmql_file", type=str, help="path to the LMQL file to run")
    parser.add_argument("--no-clear", action="store_true", dest="no_clear", help="don't clear inbetween printing results")
    parser.add_argument("--no-realtime", action="store_true", dest="no_realtime", help="don't print text as it's being generated")
    parser.add_argument("--time", action="store_true", dest="time", help="Time the query.")

    args = parser.parse_args(sys.argv[2:])

    absolute_path = os.path.abspath(args.lmql_file)
    from lmql.runtime.output_writer import PrintingDebuggerOutputWriter

    writer = PrintingDebuggerOutputWriter()
    writer.clear = not args.no_clear
    writer.print_output = not args.no_realtime

    results = asyncio.run(lmql.run_file(absolute_path, output_writer=writer))
    
    if type(results) is not list:
        results = [results]
    
    for r in results:
        for v in [v for v in r.variables if v.startswith("P(")]:
            distribution = r.variables[v]
            max_prob = max([p for _,p in distribution])
            labels = []
            for value, prob in distribution:
                label = value if prob != max_prob else f"{value} (*)"
                labels.append(label)
            max_length = max([len(str(l)) for l in labels])

            print(v)
            for (value, prob), label in zip(distribution, labels):
                label = label.ljust(max_length)
                print(" - {} {}".format(label, prob))

    if args.time:
        print("Query took:", time.time() - start)

def ensure_node_install():
    try:
        v = subprocess.check_output("node --version", shell=True, stderr=subprocess.DEVNULL).decode("utf-8").strip()
    except:
        print("""node.js is not installed. Please install it to use the LMQL playground.

If your Python installation is managed by conda, you can install node.js with:

    conda install nodejs=14.20 -c conda-forge
    
Alternatively, you can find instructions for installing node.js on your system on the official website: https://nodejs.org/en/download/.""")
        sys.exit(1)

def cmd_playground():
    """
    emoji:💻 runs LMQL in development mode (hot-reloading python and debugger implementation)
    """
    ensure_node_install()
    
    parser = argparse.ArgumentParser(description="Launches an instance of the LMQL playground.")
    parser.add_argument("--live-port", type=int, default=3004, help="port to use to host the LMQL live server")
    parser.add_argument("--ui-port", type=int, default=3000, help="port to use to host the LMQL debugger UI")
    
    args = parser.parse_args(sys.argv[2:])

    project_root = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    print(f"[lmql playground {project_root}, liveserver=localhost:{args.live_port}, ui=localhost:{args.ui_port}]")

    # make sure yarn is installed
    os.system("npm install -g yarn")

    # repo commit
    if os.path.exists(os.path.join(project_root, "../.git")):
        commit = subprocess.check_output("git rev-parse HEAD", shell=True, cwd=project_root).decode("utf-8").strip()
        commit = commit[:7]
        has_uncomitted_files = len(subprocess.check_output("git status --porcelain", shell=True, cwd=project_root).decode("utf-8").strip()) > 0
        if has_uncomitted_files:
            commit += " (dirty)"
    else:        
        commit = version_info.commit

    # live server that executes LMQL queries and returns results and debugger data
    live_process = subprocess.Popen("yarn && PORT=" + str(args.live_port) + " node live.js", 
        shell=True, cwd=os.path.join(project_root, "lmql/ui/live"))

    # UI that displays the debugger (uses live server API for data and remote execution)
    ui_modern_process = subprocess.Popen(f"yarn && REACT_APP_BUILD_COMMIT='{commit}' REACT_APP_SOCKET_PORT={args.live_port} yarn run start", 
        shell=True, cwd=os.path.join(project_root, "lmql/ui/playground"))

    try:
        live_process.wait()
        ui_modern_process.wait()
    except KeyboardInterrupt:
        print("[lmql playground] Ctrl+C pressed, exiting...")
        live_process.terminate()
        ui_modern_process.terminate()

def cmd_usage():
    """
    emoji:❓ show usage information
    """
    commands = [f for f in globals().values() if callable(f) and f.__name__.startswith("cmd_")]
    commands = [(f.__name__[4:], f.__doc__ or "<no description") for f in commands]

    def format_command_line(c, doc):
        emoji = "   "
        doc = doc.strip()
        if doc.startswith("emoji:"):
            emoji, doc = doc[len("emoji:"):].split(" ", 1)
            emoji += " "
            doc = doc.strip()
        return ("\n  " if c == "usage" else "") + emoji + c.replace("_", "-") + " " * (20 - len(c)) + " " + doc

    command_list = "\n".join(["  " + format_command_line(name, doc) for name, doc in commands])
    if version_info.commit == "dev":
        print(f"[LMQL Dev Version {project_root}]\n")
    print(f"""USAGE: lmql <command>

Commands:

{command_list}
""")


def cmd_version():
    """
    emoji:📦 show version information
    """
    print(f"lmql v{version_info.version}")
    if version_info.commit != "dev":
        print(f"  commit: {version_info.commit}")
    print(f"  path: {project_root}/lmql")
    if version_info.build_on != "dev":
        print(f"  build on: {version_info.build_on}")

def hello():
    import asyncio
    code_local = """
argmax "Hello[WHO]" from "local:gpt2-medium" where len(WHO) < 10    
"""
    print("[Greeting 🤗 Transformers]")
    asyncio.run(lmql.run(code_local))
    
    print("[Greeting OpenAI]")
    code_openai = """
argmax "Hello[WHO]" from "openai/text-ada-001" where len(WHO) < 10    
"""
    asyncio.run(lmql.run(code_openai))

hidden_commands = {
    "hello": hello
}

def main():
    if len(sys.argv) < 2:
        cmd_usage()
        sys.exit(1)
    # get all functions defined in this file
    functions = [f for f in globals().values() if callable(f)]
    # get all functions that start with "cmd_"
    commands = [f for f in functions if f.__name__.startswith("cmd_")]
    # get the command function
    command_name = sys.argv[1]
    if command_name == "dev":
        print("'lmql dev' is deprecated, use 'lmql playground' instead.")
        command_name = "playground"
    if command_name in hidden_commands.keys():
        hidden_commands[command_name]()
        sys.exit(0)
    command = [f for f in commands if f.__name__ == "cmd_" + command_name or f.__name__ == "cmd_" + command_name.replace("-", "_")]
    if len(command) == 0:
        print("Unknown command: " + sys.argv[1])
        cmd_usage()
        sys.exit(1)
    command[0]()

if __name__ == "__main__":
    main()



from typing import List, Dict, Any

# check if langchain is available
try:
    from langchain.chains.base import Chain

    class LMQLChainMixIn(Chain):
        def _call(self, inputs: Dict[str, str]) -> Dict[str, str]:
            import asyncio

            if LMQLChainMixIn.loop is None:
                LMQLChainMixIn.loop = asyncio.new_event_loop()
            loop = LMQLChainMixIn.loop
            
            asyncio.set_event_loop(loop)

            argnames = self.args
            args = [inputs[argname] for argname in argnames]

            task = self.__acall__(*args)
            try:
                res = loop.run_until_complete(task)
            except:
                LMQLChainMixIn.loop = asyncio.get_event_loop()
                print(LMQLChainMixIn.loop)
                res = loop.run_in_executor(None, task)
            res = [{k:v for k,v in r.variables.items()} for r in res]

            if len(res) == 1:
                res = res[0]

            return res

    LMQLChainMixIn.loop = None
except ImportError as e:
    class LMQLChainMixIn:
        pass


import lmql.version as version_info
from lmql.runtime.lmql_runtime import LMQLInputVariableScope

"""
lmql.

A query language for language models.
"""

__version__ = version_info.version
__author__ = 'Luca Beurer-Kellner, Marc Fischer and Mark Mueller'
__email__ = "luca.beurer-kellner@inf.ethz.ch"
__license__ = "MIT"

from lmql.language.compiler import LMQLCompiler
import lmql.runtime.lmql_runtime as lmql_runtime
import tempfile
from lmql.runtime.output_writer import silent, stream, printing
from lmql.runtime.model_registry import LMQLModelRegistry
from lmql.runtime.lmql_runtime import LMQLQueryFunction, FunctionContext

import lmql.runtime.lmql_runtime as runtime_support

# re-export lmql runtime functions
from lmql.runtime.lmql_runtime import compiled_query, tag

import os

model_registry = LMQLModelRegistry

def connect(server="http://localhost:8080", model_name="EleutherAI/gpt-j-6B"):
    from lmql.runtime.hf_integration import transformers_model

    Model = transformers_model(server, model_name)
    lmql_runtime.register_model(model_name, Model)
    lmql_runtime.register_model("*", Model)

def _autoconnect_model(model_name):
    if model_name.startswith("openai/"):
        from lmql.runtime.openai_integration import openai_model

        # hard-code openai/ namespace to be openai-API-based
        Model = openai_model(model_name)
        lmql_runtime.register_model(model_name, Model)
        lmql_runtime.register_model("*", Model)
    else:
        if "LMQL_BROWSER" in os.environ:
            assert False, "Cannot use HuggingFace Transformers models in browser. Please use openai/ models or install lmql on your local machine."

        from lmql.runtime.hf_integration import transformers_model

        default_server = "http://localhost:8080"
        Model = transformers_model(default_server, model_name)
        lmql_runtime.register_model(model_name, Model)
        lmql_runtime.register_model("*", Model)

def make_pinned_model(model_name):
    _autoconnect_model(model_name)
    return model_registry.get(model_name)

def autoconnect():
    model_registry.autoconnect = _autoconnect_model

def load(filepath=None, autoconnect=False, force_model=None, output_writer=None):
    if autoconnect: 
        model_registry.autoconnect = _autoconnect_model
    # compile query and obtain the where clause computational graph
    compiler = LMQLCompiler()
    module = compiler.compile(filepath)
    if module is None: 
        return None
    
    if output_writer is not None:
        output_writer.add_compiler_output(module.code())

    module = module.load()
    module.query.force_model(force_model)
    return module

async def run_file(filepath, output_writer=None, force_model=None, *args):
    module = load(filepath, autoconnect=True, output_writer=output_writer, force_model=force_model)
    
    if module is None: 
        print("Failed to compile query.")
        return

    if output_writer is not None:
        module.query.output_writer = output_writer
    
    if len(args) == 1 and args[0] == "":
        kwargs = {}
    else:
        kwargs = {}
        for line in args:
            line = line.strip()
            key, value = line.split(":", 1)
            kwargs[key.strip()] = value.strip()

    return await module.query(**kwargs)

async def run(code, output_writer=None):
    temp_lmql_file = tempfile.mktemp(suffix=".lmql")
    with open(temp_lmql_file, "w") as f:
        f.write(code)
    
    os.chdir(os.path.join(os.path.dirname(__file__), "../../")) 
    return await run_file(temp_lmql_file, output_writer=output_writer)
        
def _query_from_string(s):
    temp_lmql_file = tempfile.mktemp(suffix=".lmql")
    with open(temp_lmql_file, "w") as f:
        f.write(s)
    module = load(temp_lmql_file, autoconnect=True, output_writer=silent)
    return module.query
        
def query(fct):
    import inspect
    
    # support for lmql.query(<query string>)
    if type(fct) is str: return _query_from_string(fct)
    
    calling_frame = inspect.stack()[1]
    scope = LMQLInputVariableScope(fct, calling_frame)
    code = fct.__doc__
    
    temp_lmql_file = tempfile.mktemp(suffix=".lmql")
    with open(temp_lmql_file, "w") as f:
        f.write(code)
    module = load(temp_lmql_file, autoconnect=True, output_writer=silent)
    
    assert inspect.iscoroutinefunction(fct), f"@lmql.query {fct.__name__} must be declared async."
    
    argnames = inspect.getfullargspec(fct).args
    
    args_of_query = [a for a in inspect.getfullargspec(module.query.fct).args if a != "context"]
    # print code of module.query.fct
    for a in argnames:
        if a not in args_of_query:
            print(f"warning: @lmql.query {fct.__name__} has an argument '{a}' that is not used in the query.")
    
    # set the function context of the query based on the function context of the decorated function
    module.query.function_context = FunctionContext(argnames, args_of_query, scope)
    
    return module.query


import ast
import sys
import astunparse
import os
import importlib
import re

from io import StringIO
from lmql.ops.ops import lmql_operation_registry

from lmql.language.qstrings import qstring_to_stmts, TemplateVariable
from lmql.language.validator import LMQLValidator, LMQLValidationError
from lmql.language.fragment_parser import LMQLDecoderConfiguration, LMQLQuery, LanguageFragmentParser, FragmentParserError
from lmql.runtime.model_registry import model_name_aliases

OPS_NAMESPACE = "lmql.ops"

class PromptScope(ast.NodeVisitor):
    def scope(self, query: LMQLQuery):
        self.distribution_vars = set([query.distribution.variable_name]) if query.distribution is not None else set()    
        self.defined_vars = set()

        # collect set of global query template variables
        self.free_vars = set()
        self.written_vars = set()

        for p in query.prompt: self.visit(p)

        query.scope = self

    def visit_Constant(self, node):
        if type(node.value) is not str: return super().visit_Constant(node)
        qstring = node.value

        # capture set of defined vars
        declared_template_vars = [v.name for v in qstring_to_stmts(qstring) if type(v) is TemplateVariable]
        for v in declared_template_vars: 
            self.defined_vars.add(v)
            self.written_vars.add(v)
            if v in self.free_vars: self.free_vars.remove(v)

        # capture set of format vars
        used_fstring_expr = [v[1:-1] for v in re.findall("\{[^\}]+\}", qstring)]
        for v in used_fstring_expr:
            if v.startswith(":"):
                continue
            parsed = ast.parse(v).body[0].value
            self.visit(parsed)
            
            # if v not in self.defined_vars and v not in self.free_vars and v not in self.written_vars:
            #     self.free_vars.add(v)
                
        template_tags = [v[1:-1] for v in re.findall("\{:[A-z0-9]+\}", qstring)]
        for tt in template_tags:
            qstring = qstring.replace(f"{{{tt}}}", f"{{lmql.tag('{tt[1:]}')}}")
        
        node.value = qstring

        return super().visit_Constant(node)

    def visit_Name(self, node: ast.Name):
        name = str(node.id)
        
        if type(node.ctx) is ast.Store:
            self.written_vars.add(name)
            if name in self.free_vars:
                self.free_vars.remove(name)
            
        if type(node.ctx) is ast.Load:
            if name not in self.free_vars and name not in self.written_vars:
                self.free_vars.add(name)
        
        return True

class QueryStringTransformation(ast.NodeTransformer):
    """
    Transformes string expressions on statement level to model queries.
    """
    
    def __init__(self, query):
        self.query = query
    
    def transform(self):
        self.query.prompt = [self.visit(p) for p in self.query.prompt]

    def visit_Expr(self, expr):
        if type(expr.value) is ast.Constant:
            expr.value = self.transform_Constant(expr.value)
        else:
            self.generic_visit(expr)
        return expr

    def transform_Constant(self, constant):
        if type(constant.value) is not str: return constant
        qstring = constant.value
        # TODO: handle escaping more completely and gracefully
        qstring = qstring.replace("\n", "\\\\n")
        compiled_qstring = ""

        declared_template_vars = set()

        for i, stmt in enumerate(qstring_to_stmts(qstring)):
            if type(stmt) is str:
                compiled_qstring += stmt
            elif type(stmt) is TemplateVariable:
                if stmt.name in self.query.scope.distribution_vars:
                    compiled_qstring += f"[distribution:{stmt.name}]"
                else:
                    declared_template_vars.add(stmt.name)
                    compiled_qstring += "[" + stmt.name + "]"

        if len(compiled_qstring) == 0:
            return constant

        result_code = f'yield context.query(f"""{compiled_qstring}""")'

        for v in declared_template_vars:
            result_code += f"\n{v} = context.get_var('{v}')"
        return ast.parse(result_code)

    # def transform_prompt_stmt(self, stmt):
    #     if type(stmt) is ast.Expr:
    #         if type(stmt.value) is ast.Constant and type(stmt.value.value) is str:
    #             # print(stmt.value.value)
    #             stmt.value = self.transform_query_string(stmt.value.value)
    #             # print(dir(stmt.value))
    #     else:
    #         print(type(stmt))
    #     return stmt

class SNFList:
    def __init__(self):
        self.stmts = []
        self.var_counter = 0

    def add(self, expr):
        vname = f"intm{self.var_counter}"
        self.var_counter += 1
        self.stmts.append(f"{vname} = {expr.strip()}")
        return vname

    def ast(self):
        return ast.parse("\n".join(self.stmts))
    
    def last_var(self):
        assert self.var_counter > 0, "No last variable available (0 statements in SNF)."
        return f"intm{self.var_counter - 1}"

    def str(self):
        return "\n".join(self.stmts)

class NameVisitor(ast.NodeVisitor):
    def __init__(self, name_visitor):
        self.name_visitor = name_visitor

    def visit_Name(self, node):
        self.name_visitor(node)


class NameTransformer(ast.NodeTransformer):
    def __init__(self, name_transformer):
        self.name_transformer = name_transformer

    def visit_Name(self, node):
        return self.name_transformer(node)

class ReturnStatementTransformer(ast.NodeTransformer):
    def __init__(self, query):
        self.query = query
    
    def transform(self):
        self.query.prompt = [self.visit(p) for p in self.query.prompt]

    def visit_Return(self, node):
        return ast.parse("yield ('result', " + astunparse.unparse(node.value).strip() + ")")

class WhereClauseTransformation():
    def __init__(self, query: LMQLQuery):
        self.query = query
        self.scope: PromptScope = query.scope

        assert self.scope is not None, "WhereClauseTransformation requires a scoped query for transformation"
    
    def transform(self):
        snf = SNFList()
        
        if self.query.where is None:  return None
        if type(self.query.where) is ast.Expr: 
            self.query.where = self.query.where.value

        result = self.transform_node(self.query.where, snf=snf)

        self.query.where = snf.str()
        self.query.where_expr = result

    def transform_name(self, node, keep_variables=False, plain_python=False):
        # check for built-ins
        bn = get_builtin_name(node, plain_python)
        if bn is not None: return bn
        # check if variable is distribution variable
        if node.id in self.scope.distribution_vars:
            raise LMQLValidationError("Distribution variable {} cannot be used in where clause.".format(node.id))

        # check for template variables
        if node.id in self.scope.defined_vars and not keep_variables:
            return f"lmql.runtime_support.Var('{node.id}')"

        return bn or node.id

    def transform_node(self, expr, snf):
        if type(expr) is ast.BoolOp:
            if type(expr.op) is ast.And or type(expr.op) is ast.Or:
                ops = expr.values
                tops = [self.transform_node(op, snf) for op in ops]
                tops_list = ",\n  ".join([t.strip() or "None" for t in tops])
                
                Op = "lmql.runtime_support.AndOp" if type(expr.op) is ast.And else "lmql.OrOp"
                return snf.add(f"{Op}([\n  {tops_list}\n])")
        # elif type(expr) is ast.Call:
        #     tfunc = self.transform_node(expr.func, snf)
        #     targs = [self.transform_node(a, snf) for a in expr.args]
        #     targs_list = ", ".join(targs)
        #     return f"{tfunc}({targs_list})"
        elif type(expr) is ast.Name:
            return self.transform_name(expr)
        elif type(expr) is ast.UnaryOp:
            op = expr.op
            
            Ops = {
                ast.Not: f"{OPS_NAMESPACE}.NotOp"
            }

            for OpT, impl in Ops.items():
                if type(op) is OpT:
                    operand = self.transform_node(expr.operand, snf).strip()
                    return snf.add(f"{impl}([{operand}])")
            
            assert False, "unary operator {} not supported.".format(type(expr.op))
        elif type(expr) is ast.Compare:
            op = expr.ops[0]
            assert len(expr.ops) == 1, "compiler currently does not support comparison with more than one operator"
            
            Ops = {
                ast.Eq: f"{OPS_NAMESPACE}.EqOp",
                ast.Lt: f"{OPS_NAMESPACE}.Lt",
                ast.Gt: f"{OPS_NAMESPACE}.Gt",
                ast.In: f"{OPS_NAMESPACE}.InOp"
            }

            for OpT, impl in Ops.items():
                if type(op) is OpT: 
                    ops = [self.transform_node(c, snf) for c in [expr.left] + expr.comparators]
                    ops_list = ", ".join(ops).strip()
                    return snf.add(f"{impl}([{ops_list}])")
            
            # if is_type_constraint(expr):
            #     type_name = expr.comparators[0].id
            #     var_name = expr.left.args[0].id
            #     return snf.add(f"{OPS_NAMESPACE}.TypeConstraint([{type_name}, {OPS_NAMESPACE}.Var('{var_name}')])")
            
            assert False, "operator {} is not supported.".format(astunparse.unparse(expr))
        elif type(expr) is ast.Constant:
            return self.default_transform_node(expr, snf).strip()
        elif type(expr) is ast.ListComp:
            return self.default_transform_node(expr, snf).strip()
        elif type(expr) is ast.Call:
            bn = get_builtin_name(expr.func)
            if bn is not None:
                args = [self.transform_node(a, snf) for a in expr.args]
                args_list = ", ".join(args)
                return f"{bn}([{args_list}])"
            if is_allowed_builtin_python_call(expr.func):
                return self.default_transform_node(expr, snf).strip()
        elif type(expr) is ast.List:
            return self.default_transform_node(expr, snf).strip()

        print(f"compiler warning: expressions of type {type(expr)} are not explicitly supported: '{astunparse.unparse(expr).strip()}'")
        return snf.add(self.default_transform_node(expr, snf))

    def default_transform_node(self, node, snf):
        # collect the set of captured template variables
        names = set()
        def collect_name(node):
            name = node.id
            if name in self.scope.defined_vars:
                names.add(name)
        NameVisitor(collect_name).visit(node)
        names = sorted(list(names))

        # if no template variable names are capture, node represents a constant expression
        if len(names) == 0: return astunparse.unparse(node)

        def transform_name(node):
            node.id = self.transform_name(node, keep_variables=True, plain_python=True)
            return node
        node = NameTransformer(transform_name).visit(node)

        args = (" " + ", ".join(names)) if len(names) > 0 else ""
        var_ops = (", ".join([f"{OPS_NAMESPACE}.Var('{n}')" for n in names])).strip() if len(names) > 0 else ""
        fct_code = astunparse.unparse(node).strip()
        
        return f"{OPS_NAMESPACE}.OpaqueLambdaOp([lambda{args}: {fct_code}, {var_ops}])"

def is_allowed_builtin_python_call(node):
    if type(node) is not ast.Name:
        return False
    allowed_builtin_functions = set(["set", "all"])
    return node.id in allowed_builtin_functions

def get_builtin_name(node, plain_python=False):
    if type(node) is not ast.Name:
        return None
    n = node.id
    
    if n in lmql_operation_registry.keys():
        return OPS_NAMESPACE + "." + lmql_operation_registry[n]
    
    return None

def is_type_constraint(expr: ast.Expr):
    if not type(expr.left) is ast.Call:
        return False
    if not type(expr.left.func) is ast.Name:
        return False
    if not expr.left.func.id == "type":
        return False
    if not type(expr.ops[0]) is ast.Is:
        return False
    if not len(expr.comparators) == 1:
        return False
    right = expr.comparators[0]
    if not type(right) is ast.Name:
        return False
    if not len(expr.left.args) == 1:
        return False
    if not type(expr.left.args[0]) is ast.Name:
        return False
    
    return True

class DecodeClauseTransformation:
    def __init__(self, query):
        self.query = query

    def transform(self):
        if type(self.query.decode) is ast.Name:
            method = ast.Constant(self.query.decode.id, "str")
            keyword_args = []
            self.query.decode = LMQLDecoderConfiguration(method, keyword_args)
        elif type(self.query.decode) is ast.Call:
            method = ast.Constant(self.query.decode.func.id, "str")
            keyword_args = self.query.decode.keywords
            self.query.decode = LMQLDecoderConfiguration(method, keyword_args)
        else:
            assert False, "cannot handle decode clause {} ()".format(self.query.decode, type(self.query.decode))


        return self.query

class CompilerTransformations:
    def __init__(self):
        self.transformations = [
            QueryStringTransformation,
            WhereClauseTransformation,
            DecodeClauseTransformation,
            ReturnStatementTransformer
        ]
    
    def transform(self, query):
        for T in self.transformations:
            t = T(query).transform()
        return query

class PythonFunctionWriter:
    def __init__(self, name, filename, parameters, prologue, decorators=None, decorators_args=None):
        self.name = name
        self.filename = filename
        self.parameters = parameters
        self.prologue = prologue
        
        self.in_memory_contents = ""
        
        self.file = open(self.filename, "w")

        self.indent = "  "
        self.write("import lmql\n")
        self.write(self.prologue)
        
        if decorators is not None:
            if decorators_args is None:
                decorators_args = [None] * len(decorators)
            for d,args in zip(decorators, decorators_args):
                self.write(f"@{d}({args})\n")
        self.write(f"async def {self.name}({self.make_kwargs()}):\n")

    def write(self, code):
        self.in_memory_contents += code
        self.file.write(code)

    def make_kwargs(self):
        if len(self.parameters) == 0:
            return ""
        return ",".join(f"{p}=None" for p in self.parameters)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.file.close()

    def add(self, code):
        if code is None:
            return
        for line in code.split("\n"):
            if line.strip() == "": continue
            self.write(f"{self.indent} {line}\n")

class LMQLModule(object):
    def __init__(self, compiled_file, lmql_code=None, output_variables=None):
        self.compiled_file = compiled_file
        self._code = None
        self.lmql_code = lmql_code
        self.output_variables = output_variables or []

    def load(self):
        sys.path.append(os.path.dirname(self.compiled_file))
        m = __import__(os.path.basename(self.compiled_file[:-3]))
        setattr(m, "code", self.code)
        setattr(m, "lmql_code", self.lmql_code)
        return m

    def __str__(self):
        with open(self.compiled_file, "r") as f:
            return f.read()

    def code(self):
        if self._code is None:
            with open(self.compiled_file, "r") as f:
                self._code = f.read()
        return self._code

def unparse_list(ast_elements):
    return ", ".join([astunparse.unparse(e).strip() for e in ast_elements])

def preprocess_text(lmql_code):
    if lmql_code.startswith("lmql"):
        lmql_code = lmql_code[4:]
    while lmql_code.startswith("\n") or lmql_code.startswith(" "):
        if lmql_code.startswith("\n"):
            lmql_code = lmql_code[1:]
            break
        else:
            lmql_code = lmql_code[1:]
    
    # remove common indent
    lines = lmql_code.split("\n")
    common_indent = min([len(l) - len(l.lstrip()) for l in lines if len(l.strip()) > 0])
    return "\n".join([l[common_indent:] for l in lines])

class LMQLCompiler:
    def __init__(self):
        pass

    def compile(self, filepath):
        try:
            # parse file
            with open(filepath) as f:
                contents = f.read()
            lmql_code = preprocess_text(contents)
            buf = StringIO(lmql_code)
            parser = LanguageFragmentParser()
            q = parser.parse(buf.readline)

            # output file path
            basename = os.path.basename(filepath).split(".lmql")[0]
            output_file = os.path.join(os.path.dirname(filepath), f"{basename}_compiled.py")

            # scoping
            scope = PromptScope()
            scope.scope(q)

            # validation
            LMQLValidator().validate(q)

            # compilation
            transformations = CompilerTransformations()
            transformations.transform(q)

            model_name = astunparse.unparse(q.from_ast).strip()
            model_name = model_name_aliases.get(model_name, model_name)
            if model_name[1:-1] in model_name_aliases.keys():
                model_name = "'" + model_name_aliases[model_name[1:-1]] + "'"

            # resulting code
            code = None
            output_variables = "output_variables=[" + ", ".join([f'"{v}"' for v in scope.defined_vars]) + "]"

            # generate function that runs query
            with PythonFunctionWriter("query", output_file, list(scope.free_vars) + ["context"], 
                q.prologue, decorators=["lmql.compiled_query"], decorators_args=[output_variables]) as writer:
                
                writer.add(f"context.set_model({model_name})")
                writer.add(f"context.set_decoder({astunparse.unparse(q.decode.method).strip()}, {unparse_list(q.decode.decoding_args)})")
                writer.add("# where")
                writer.add(q.where)
                writer.add(f"context.set_where_clause({q.where_expr})")
                writer.add("# prompt")
                writer.add(astunparse.unparse(q.prompt))
                if q.distribution:
                    writer.add("# distribution")
                    writer.add("context.set_distribution('{}', {})".format(q.distribution.variable_name, astunparse.unparse(q.distribution.values).strip()))
                
                writer.add(f"yield ('result', context.get_return_value())")

            return LMQLModule(output_file, lmql_code=lmql_code, output_variables=[v for v in scope.defined_vars])
        except FragmentParserError as e:
            sys.stderr.write("error: " + str(e) + "\n")




from lmql.runtime.decoder_head import DecoderHead
from lmql.runtime.prompt_interpreter import ProgramState

import numpy as np
from lmql.utils import nputil

class ScoringQuery:
    def __init__(self, result, scoring_head_index, prompt, values, output_writer):
        self.variable_name = result.distribution_variable
        self.result = result
        self.scoring_head_index = scoring_head_index

        self.prompt = prompt
        self.values = values
        self.output_writer = output_writer
        
        self.value_input_ids = None
        self.scores = [[] for i in range(len(values))]
        
        self.batch_idx = 0
        self.batch_size = -1

        self.log_score_prompt = False

    async def where(self, head: DecoderHead):
        if self.value_input_ids is None:
            self.value_input_ids = [await head.tokenize(v) for v in self.values]

        idx = self.batch_idx * self.batch_size + head.seq_idx 

        next_token_logits = head.next_token_logits
        next_token_logits = np.zeros_like(next_token_logits)
        
        if len(self.value_input_ids[idx]) == 0:
            head.next_token_logits[:] = np.finfo(np.float32).min
            next_token_logits[head.eos_token_id] = 100
            return next_token_logits

        head.next_token_logits[:] = np.finfo(np.float32).min
        next_token_logits[self.value_input_ids[idx][0]] = 100
        self.value_input_ids[idx] = self.value_input_ids[idx][1:]

        return next_token_logits
    
    async def rewriter(self, head: DecoderHead):
        score = nputil.log_softmax(head.next_token_logits)[head.next_token_id]
        
        idx = self.batch_idx * self.batch_size + head.seq_idx 

        if head.next_token_id != head.eos_token_id:
            self.scores[idx] += [score]
            await self.debugger_output(head)
        
        if self.log_score_prompt:
            text = await head.detokenize(head.input_ids_without_padding.reshape(-1).tolist())
            print("Scored", "\n".join(l for l in text.split("\n")[-5:]), "with", sum(self.scores[idx] + [0]))
        
        return None

    async def debugger_output(self, head):
        if self.output_writer is None: return

        idx = self.batch_idx * self.batch_size + head.seq_idx
        # more expensive but needed when there are bugs in input_id rewriting
        full_ids = head.input_ids_without_padding.reshape(-1).tolist() + [head.next_token_id]
        full_text = await head.detokenize(full_ids)
        
        value = self.values[idx]

        program_state = ProgramState()
        program_state.variable_values = self.result.variables.copy()
        program_state.variable_values[self.variable_name] = value
        program_state.variable_values[f"score({self.variable_name})"] = sum(self.scores[idx] + [0])

        self.output_writer.add_interpreter_head_state("P({} = \"{}\")".format(self.variable_name, value), self.scoring_head_index, 
            full_text, None, None, True, "fin", None, len(full_ids), program_state)


    async def score(self, model, batch_size=None):
        self.batch_size = batch_size
        if self.batch_size is None:
            self.batch_size = len(self.values)
        
        self.scores = await model.score_distribution_values(self.prompt, self.values)
        return [np.array(s) for s in self.scores]

class ConditionalDistributionPostprocessor:
    def __init__(self, interpreter):
        self.interpreter = interpreter
        self.output_writer = interpreter.output_writer

    async def process(self, results):
        model = self.interpreter.model

        if type(results) is not list:
            results = [results]

        # check if distribution is required
        if not any(r is not None and hasattr(r, "distribution_variable") and r.distribution_variable is not None for r in results):
            return results

        if len(results) > 1:
            if "top1_distribution" in self.interpreter.specified_decoder_args and self.interpreter.specified_decoder_args["top1_distribution"]:
                print("top1_distribution: only computing conditional distribution for the top1 result")
                results = [results[0]]
            else:
                print("warning: more than one result, computing conditional distribution for all of them.")

        for i, result in enumerate(results):
            if result is None: continue

            distribution_variable = result.distribution_variable
            distribution_values = result.distribution_values

            prompts = [result.prompt + value for value in distribution_values]

            if distribution_variable is None:
                print("warning: result {} has no distribution variable set even though a DISTRIBUTION clause is given".format(i))
                continue

            scorer = ScoringQuery(result, i, result.prompt, distribution_values, self.output_writer)
            scores = await scorer.score(model, batch_size=self.interpreter.specified_decoder_args.get("distribution_batch_size", None))

            # print("Computing P({} | {}) for result {}...".format(distribution_variable, result.prompt[:10] + "...", i))

            scores = np.stack([s.sum() for s in scores], axis=0)
            log_probs = nputil.log_softmax(scores)
            probs = np.exp(log_probs)
            
            distribution = [(value, prob, prompt) for value, prob, prompt in zip(distribution_values, probs, prompts)]
            log_distribution = [(value, log_prob, prompt) for value, log_prob, prompt in zip(distribution_values, log_probs, prompts)]

            result.variables[distribution_variable] = max(distribution, key=lambda x: x[1])[0]
            result.prompt = max(distribution, key=lambda x: x[1])[2] # get prompt including distribution value for argmax result
            
            result.variables[f"P({distribution_variable})"] = [(value, prob) for value, prob, _ in distribution]
            result.variables[f"log P({distribution_variable})"] = [(value, prob) for value, prob, _ in log_distribution]

        if len(results) == 1:
            return results[0]
        else:
            return results



import os

import sys
sys.path.append("../")
sys.path.append(os.path.join(os.path.dirname(__file__), "../../"))

from livelib import LiveApp, live, add_debugger_output
import json
import numpy as np
import asyncio

class LiveDebuggerOutputWriter:
    def __init__(self, web):
        self.records_graph = True
        self.web = web

    def add_decoder_state(self, graph_dict):
        add_debugger_output("decoder-graph-state", graph_dict)
    
    def report_model_stats(self, **kwargs):
        add_debugger_output("openai-token-count", kwargs)
        
    async def input(self, *args):
        add_debugger_output("stdin-request", {})
        return await LiveApp.ainput(*args, web=self.web)

    def add_interpreter_head_state(self, variable, head, prompt, where, trace, is_valid, is_final, mask, num_tokens, program_variables):
        from lmql.utils.graph import CytoscapeGraphWriter
        
        def node_data(op):
            result = "-"
            if trace is not None and op in trace:
                result = trace[op]

            follow_map = "-"
            if hasattr(op, "follow_map"):
                follow_map = str(op.follow_map)
            return {
                "result": result,
                "follow_map": follow_map,
                "repr": repr(op)
            }

        writer = CytoscapeGraphWriter(extra_data_provider=node_data)
        writer.write(where)

        args = ("interpreter-head", {
            "variable": variable,
            "head_index": head,
            "prompt": prompt,
            "mask": str(mask),
            "valid": str(is_valid),
            "final": is_final,
            "num_tokens": num_tokens,
            "program_variables": program_variables.variable_values if program_variables is not None else {},
            "where": writer.graph.to_json()
        })
        add_debugger_output(*args)
        
    def add_compiler_output(self, code): 
        add_debugger_output("compiled-code", {
            "code": code
        })

@live()
async def lmql(code, *args, web=False):
    import lmql
    
    if code.startswith("./"):
        with open(code) as f:
            code = f.read()

    output_writer = LiveDebuggerOutputWriter(web=web)

    result = await lmql.run(code, output_writer=output_writer)

    for r in (result if type(result) is list else [result]):
        if r is None:
            continue
        
        for v in [v for v in r.variables if v.startswith("P(")]:
            distribution = r.variables[v]
            max_prob = max([p for _,p in distribution])
            labels = []
            for value, prob in distribution:
                label = value if prob != max_prob else f"{value} (*)"
                labels.append(label)
            max_length = max([len(str(l)) for l in labels])

            print(v)
            for (value, prob), label in zip(distribution, labels):
                label = label.ljust(max_length)
                print(" - {} {}".format(label, prob))

def replace_inf_nan_with_str(d):
    import math
    for k, v in d.items():
        if type(v) is dict:
            replace_inf_nan_with_str(v)
        elif type(v) is float:
            if math.isinf(v) or math.isnan(v):
                d[k] = str(v)
    return d

if __name__ == "__main__":
    LiveApp.cli()



from lmql.ops import FollowMap, fmap

def test_overlapping_product():
    f1 = fmap(
        ("a", 1),
        ("b", 2),
        ("*", 3),
    )
    f2 = fmap(
        ("b", 1),
        ("c", 2),
        ("*", 3),
    )

    # print(f1.product(f2))
    assert f1.product(f2) == fmap(
        ("a", (1,3)),
        ("b", (2,1)),
        ("c", (3,2)),
        ("*", (3,3))
    )

def test_disjunct_product():
    f1 = fmap()
    f2 = fmap()

    assert str(f1.product(f2)) == "∅"

def test_disjunct_product2():
    f1 = fmap()
    f2 = fmap(
        ("a", "abc")
    )

    assert str(f1.product(f2)) == "∅"

def test_disjunct_product3():
    f1 = fmap(
        ("b", "bcd")
    )
    f2 = fmap(
        ("a", "abc"),
        ("*", "fallback")
    )

    assert f1.product(f2) == fmap(
        ("b", ("bcd", "fallback"))
    )

def test_simple_case():
    f1 = fmap(
        ("a", True),
        ("b", False)
    )
    f2 = fmap(
        ("a", True),
        ("b", False)
    )

    assert f1.product(f2) == fmap(
        ("a", (True, True)),
        ("b", (False, False)),
    )

def test_three_operands():
    f1 = fmap(
        ("a", True),
        ("b", False)
    )
    f2 = fmap(
        ("a", True),
        ("b", False)
    )
    f3 = fmap(
        ("a", True),
        ("b", False)
    )

    assert f1.product(f2, f3) == fmap(
        ("a", (True, True, True)),
        ("b", (False, False, False)),
    )

if __name__ == "__main__":
    test_overlapping_product()
    test_disjunct_product()
    test_disjunct_product2()
    test_disjunct_product3()
    test_simple_case()
    test_three_operands()


from docutils import nodes
from docutils.parsers.rst import Directive
from pygments import highlight
from pygments.lexers import PythonLexer
from pygments.token import Name, Keyword
from pygments.formatters import HtmlFormatter

import os
import json

class LmqlLexer(PythonLexer):
    EXTRA_KEYWORDS = set((
        "BEAM",
        "beam",
        "ARGMAX",
        "argmax",
        "SAMPLE",
        "BEST_K",
        "best_k",
        "BEAM_VAR",
        "beam_var",
        "VAR",
        "var",
        "sample",
        "FROM",
        "from",
        "WHERE",
        "where",
        "DISTRIBUTION",
        "distribution")
    )

    def get_tokens_unprocessed(self, text):
        for index, token, value in PythonLexer.get_tokens_unprocessed(self, text):
            if token is Name and value in self.EXTRA_KEYWORDS:
                yield index, Keyword.Pseudo, value
            else:
                yield index, token, value
        
def output_format(output):
    print([output])

    variable_name = None
    result = ""
    is_escape = False
    offset = -1

    var_ids = {}
    var_ctr = 1

    while offset < len(output) - 1:
        offset += 1

        c = output[offset]

        if is_escape and (c == "[" or c == "]"):
            result += c
            is_escape = False
            offset += 1
            continue
        is_escape = False

        if c == "\\":
            is_escape = True
            continue

        if c == "[" and variable_name is None:
            start_offset = offset + 1
            print("parsing variable name", output[start_offset:], flush=True)
            while output[offset] != " " and output[offset] != "]":
                offset += 1
            variable_name = output[start_offset:offset]
            if variable_name not in var_ids:
                var_ids[variable_name] = var_ctr
                var_ctr += 1
            var_id = var_ids[variable_name]

            result += "<span class=\"variable val" + str(var_id) + "\"><span class=\"variable-name\">" + variable_name + "</span>"
            offset -= 1
            continue
        
        if c == "]" and variable_name is not None:
            result += "</span>"
            variable_name = None
            continue
        result += c
    
    return result

class LmqlSnippet(Directive):
    has_content = True

    def run(self):
        code = []
        output = []
        is_output = False
        name = None

        for l in self.content:
            if l.startswith("name::"):
                name = l[6:]
            elif l.startswith("model-output::"):
                is_output = True
            else:
                if is_output: output += [l]
                else: code += [l]
        
        code = "\n".join(code)
        original_code = code
        
        assert name is not None, "name:: is required for lmql directive in {}".format(self.state.document.current_source)

        # compute snippet_id
        outdir = self.state.document.settings.env.app.outdir
        # make sure doc-snippets exists
        if not os.path.exists(os.path.join(outdir, "doc-snippets")):
            os.makedirs(os.path.join(outdir, "doc-snippets"))
        snippet_id = os.path.relpath(self.state.document.current_source, self.state.document.settings.env.srcdir)
        # replace all but A-z0-9 with -
        snippet_id = "".join([c if c.isalnum() else "-" for c in snippet_id])
        snippet_id += "-" + name

        # create output
        paragraph_node = nodes.paragraph()
        prefix = """<button href onclick="openPlaygroundSnippet(this, 'doc-snippets/{}')">Open In Playground</button>""".format(snippet_id)

        code = highlight(code, LmqlLexer(), HtmlFormatter(cssclass="highlight lmql"))
        code = code.replace("""<div class="highlight lmql">""", """<div class="highlight lmql">""" + prefix)
        
        # model output highlight
        if len(output) > 0:
            print(output_format(output), flush=True)
            output = "\n".join(output)
            output_template = "<div class=\"highlight-model-output notranslate\"><div class=\"highlight\">{}</div></div>".format(output_format(output))
            code += output_template
        
        paragraph_node += nodes.raw('', code, format='html')
        
        # save snippet separately
        snippet_path = os.path.join(outdir, "doc-snippets", snippet_id + ".json")

        with open(snippet_path, "w") as f:
            playground_data = {
                "lmql-editor-contents": original_code.strip(),
                "decoder-graph": json.dumps({
                    "nodes": [],
                    "edges": []
                })
            }
            f.write(json.dumps(playground_data))

        return [paragraph_node]

def setup(app):
    app.add_directive("lmql", LmqlSnippet)
    app.add_css_file('css/lmql-docs.css')
    app.add_js_file('js/lmql-playground.js')

    return {
        'version': '0.1',
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    } 

if __name__ == "__main__":
    s = """Review: We had a great stay. Hiking in the mountains was fabulous and the food is really good.⏎', 'Q: What is the underlying sentiment of this review and why?⏎', 'A: [ANALYSIS The underlying sentiment of this review is positive because the reviewer had a great stay, enjoyed the hiking and found the food to be good.]⏎', 'Based on this, the overall sentiment of the message can be considered to be [CLASSIFICATION]"""
    print(output_format(s))


import asyncio

class InterpreterReturnValue:
    def __init__(self, value):
        self.value = value

class InterpreterCall:
    def __init__(self, *args, loc=None):
        self.args = args
        self.loc = loc

class InterpreterHeadPool:
    def __init__(self, fct, context, trace=None):
        self.heads = [InterpretationHead(fct, context, trace=trace)]

        self.done = False
        self.result = [None]

    async def advance(self, handler):
        new_heads = []
        
        for h in self.heads:
            args = await h.continue_()
            if h.done: 
                new_heads += [h]
                continue
            result = await handler(*args)
            new_heads += await h.advance(result)
        
        self.heads = new_heads[:5]
        
        self.result = [h.result for h in self.heads]
        self.done = all(h.done for h in self.heads)


class ContinueOnly: pass

def deepcopy(v):
    if isinstance(v, list):
        return [deepcopy(i) for i in v]
    elif isinstance(v, dict):
        return {k: deepcopy(v) for k,v in v.items()}
    else:
        return v

class ReplayingContextWrapper(object):
    """
    Returns the cached result (replays) the first len(self.value_cache) calls to 
    the wrapped object's methods, if they occur in self.cached_attrs.
    """

    def __init__(self, wrappee, cached_attrs=None, cache=None):
        self.wrappee = wrappee

        self.cached_attrs = cached_attrs or set()
        self.cached_attrs.add("get_var")
        self.num_replays = 0
        
        self.value_cache = []
        if cache is not None:
            self.value_cache = cache
            self.replaying = True
        else:
            self.replaying = False
        self.replay_counts = {}

    def copy(self):
        return ReplayingContextWrapper(self.wrappee, self.cached_attrs, cache=deepcopy(self.value_cache))

    def __getattr__(self, attr):
        if attr in ["cached_attrs", "value_caches", "replaying", "wrappee", "copy"]:
            return super().__getattr__(attr)

        if not attr in self.cached_attrs:
            return getattr(self.wrappee, attr)
        
        def fct(*args, **kwargs):
            # replace all cached calls
            if self.replaying:
                if len(self.value_cache) <= self.num_replays:
                    self.replaying = False
                else:
                    value = self.value_cache[self.num_replays]
                    self.num_replays += 1
                    # print("replaying {} call to".format(self.num_replays), attr, "with", args, kwargs, " as", type(value), [value])
                    return value
            
            # otherwise keep recording
            result = getattr(self.wrappee, attr)(*args, **kwargs)
            self.value_cache += [result.copy() if hasattr(result, "copy") else result]
            # print("value_cache", self.value_cache)
            return result
        return fct

class InterpretationHeadDone(Exception):
    pass

class InterpretationHead:
    def __init__(self, fct, context, args=None, kwargs=None, trace=None):
        self.fct = fct
        self.context = context
        self.context_wrapped = ReplayingContextWrapper(context)
        self.user_state = None
        
        # additional function arguments
        self.args = args
        if self.args is None: self.args = ()
        self.kwargs = kwargs
        if self.kwargs is None: self.kwargs = {}
        
        # to be given at instantiation
        self.trace = [] if trace is None else trace
        # to be recorded by this head
        self._internal_trace = []

        self.iterator_fct = None
        self.queue = []

        self.done = False
        self.result = None

        self.last_call_args = None
        self.waiting = True

        self.id = "h"
        self.copies = 0

    def get_context(self):
        self.context_wrapped.wrappee = self.context
        return self.context_wrapped

    def copy(self, next_value=None, context=None):
        # construct trace for copy
        if len(self._internal_trace) > 0:
            trace = self._internal_trace.copy()
        else:
            trace = self.trace.copy()
        
        if next_value is not None: 
            trace.append(next_value)
        if not self.waiting and (len(trace) == 0 or trace[-1] is not ContinueOnly):
            trace.append(ContinueOnly)
        # construct new head
        if context is None: 
            context = self.context
        
        c = InterpretationHead(self.fct, self.context, args=self.args, kwargs=self.kwargs, trace=trace)
        c.context_wrapped = self.context_wrapped.copy()
        c.id = self.id + "_" + str(self.copies)
        c.user_state = self.user_state
        self.copies += 1

        c.waiting = self.waiting

        return c

    @property
    def num_calls(self):
        return len(self._internal_trace)

    @property
    def num_all_calls(self):
        return len(self.trace) + len(self._internal_trace)

    async def fast_forward(self):
        """
        Advances a new interpretation head according to self.trace, 
        by fulfilling all yield calls using the values in trace.
        """
        if self.iterator_fct is None:
            self.iterator_fct = self.fct(*self.args, **self.kwargs, context=self.get_context())

        while len(self.trace) > 0:
            await self.continue_()
            v = self.trace[0]
            self.trace = self.trace[1:]
            if v is ContinueOnly:
                assert len(self.trace) == 0, "ContinueOnly must be the last element in trace."
                break
            await self.advance(v)

    async def advance(self, result):
        await self.materialize_copy_if_necessary()
        
        # guard against advancing a finished head
        if self.done: raise InterpretationHeadDone()

        assert not self.done, "Cannot advance() a finished interpreter head (check self.done)."
        assert not self.waiting, "Cannot advance() a non-waiting interpreter head. Must call continue_ first."

        self.last_call_args = None
        self.waiting = True

        self._internal_trace += [result]
        try:
            self.queue.append((await self.iterator_fct.asend(result)))
        except StopAsyncIteration as e:
            raise InterpretationHeadDone(e)

        return [self]

    async def materialize_copy_if_necessary(self):
        # materialize lazy copy if not yet done
        if self.iterator_fct is None:
            self.iterator_fct = self.fct(*self.args, **self.kwargs, context=self.get_context())

            if len(self.trace) > 0:
                await self.fast_forward()

    async def continue_(self):
        await self.materialize_copy_if_necessary()
        
        if len(self.queue) == 0:
            self.queue.append(await self.iterator_fct.__anext__())

        value = self.queue[0]
        self.queue = self.queue[1:]

        if type(value) is InterpreterReturnValue:
            self.done = True
            self.result = value.value
            return None
        elif type(value) is tuple and len(value) == 2 and value[0] == "result":
            # consider ("result", <expr>) yields as return statements
            # this avoids a dependency on this file from LMQL compiled modules
            self.done = True
            self.result = value[1]
            return None
        else:
            self.last_call_args = value.args
            self.waiting = False
            
            return self.last_call_args

class BranchingPoint:
    def __init__(self, values):
        self.values = values

class MultiHeadInterpretedFunction:
    def __init__(self, fct):
        self.fct = fct
        self.ctr = 0

    async def __call__(self, *args, **kwargs):
        pool = InterpreterHeadPool(self.fct, self)
        while not pool.done:
            await pool.advance(self.do_query)
        for i,r in enumerate(pool.result):
            print(f"head {i} result: ", r)
        return pool.result

    async def do_query(self, prompt):
        return BranchingPoint([
            "0","1"
        ])

    def query(self, prompt, loc):
        return InterpreterCall(prompt, loc=loc)

def multi_head(fct):
    return MultiHeadInterpretedFunction(fct)

@multi_head
async def fun(context):
    all_values = []
    a = yield context.query("Query A", loc="a")
    all_values.append(a)
    
    for i in range(5):
        b = yield context.query("Query B", loc="b")
        all_values.append(b)
    
    yield InterpreterReturnValue(all_values)



from lmql.ops.ops import OneOf, NextToken
from lmql.tests.expr_test_utils import run_all_tests
from lmql.ops.follow_map import fmap

def test_one_of():
    one_of = OneOf([])

    assert str(one_of.follow("", set(["abcde", "efgh"]))) == \
        "{abcde, efgh} -> True and * -> False"
    assert str(one_of.follow(NextToken, set(["abcde", "efgh"]))) == \
        "{abcde, efgh} -> True and * -> False"

    assert str(one_of.follow("ab" + NextToken, set(["abcde", "efgh"]))) == \
        "{cde} -> True and * -> False"


run_all_tests(globals())


import asyncio
import numpy as np
from typing import List, Any, Union, Optional, Dict

from lmql.runtime.tokenizer import load_tokenizer
from lmql.runtime.dclib.dclib_array import DataArray, sum_scorer, alpha_length_normalized, alpha_length_normalized_det
from lmql.runtime.dclib.dclib_seq import next_is_deterministic
import lmql.runtime.dclib as dc
from lmql.runtime.stats import Stats

@dc.decoder
async def argmax(prompt_ids: np.ndarray, n=1, max_len=2048, **kwargs):
    model = dc.model(**kwargs)
    h = dc.seqs([dc.seq(prompt_ids)] * n)
    done = dc.seqs()
    
    step = 0
    
    # provide early first result to user
    yield h

    while len(h) > 0:
        h = h.extend(await model.argmax(h))
        h = await model.rewrite(h)
        h, done = (h + done).separate_by(dc.logical_not(dc.eos), dc.lt(max_len))
        
        step += 1
        
        yield (h, done)

    dc.finish(done)

@dc.decoder
async def sample(prompt_ids: np.ndarray, temperature=1, n=1, max_len=2048, **kwargs):
    model = dc.model(**kwargs)
    h = dc.seqs([dc.seq(prompt_ids)] * n)
    done = dc.seqs()

    while len(h) > 0:
        h = h.extend(await model.sample(h, temperature=temperature))
        h = await model.rewrite(h)
        h, done = (h + done).separate_by(dc.logical_and(dc.logical_not(dc.eos), dc.lt(max_len)))

        yield (h, done)
    
    yield (h, done)


@dc.decoder
async def best_k(prompt_ids: np.ndarray, k=4, budget=30, kappa=1, beta=0.5, gamma=0.05, max_heap_size=500, top_k_continuations=5, max_len=None, **kwargs):

    class BestKScorer(dc.topk_scorer):
        def __init__(self, base_scorer, t=0, kappa=1, beta=0.5, *args, **kwargs):
            self.t = t
            self.kappa = kappa
            self.beta = beta
            self.base_scorer = base_scorer

        def __call__(self, scores, s):
            def decay(time):
                return -self.kappa*(self.t - time)**self.beta
            time = s.data('time')
            if time is None: time = -1
            return self.base_scorer(scores)+decay(time)

    def prefer_older(s, t):
        if s is None: return False 
        if t is None: return True 
        st = s.data('time')
        tt = s.data('time')
        if st is None: return False
        if tt is None: return True
        if st < tt: return True

    model = dc.model(**kwargs)
    # keep track of active beams and finished sequences
    h = dc.seqs([dc.seq(prompt_ids)] * 1)
    h.data('time', -1)
    done = dc.seqs()

    t = 0
    T = 0

    base_scorer = dc.alpha_length_normalized()
    while T < budget and len(h) > 0 or len(done) == 0:
        scorer = BestKScorer(base_scorer, t, kappa, beta)
        H, h = dc.seperate_topk(h, k, scorer=scorer)
        g = len(H)
        H = H.extend(await model.topk_continuations(H, k=top_k_continuations))
        H = await model.rewrite(H)
        not_done = dc.logical_and(dc.logical_not(dc.eos), dc.lt(max_len))
        H, H_done = H.separate_by(not_done)
        # print(len(H), len(H_done))
        done = (done + H_done)
        H = H.filter(lambda s: np.exp(s.logprobs[-1]) >= gamma)
        H.data('time', t)
        h = dc.token_unique((h + H), prefer=prefer_older)
        h = dc.topk(h, max_heap_size, scorer=scorer)

        t += 1
        T += g
        yield (h, done)
    
    done = dc.topk(done, len(done))
    done.name("final result")
    dc.finish(done)

@dc.decoder
async def beam_sample(prompt_ids: np.ndarray, n=4, max_len=None, temperature=None, **kwargs):
    n = kwargs.get("num_beams", n)
    max_len = max_len or 2048
    model = dc.model(**kwargs)

    # keep track of active beams and finished sequences
    h = dc.seqs([dc.seq(prompt_ids)] * 1)
    done = dc.seqs()

    # stopping criteria for the beam search
    not_done = dc.logical_and(dc.logical_not(dc.eos), dc.lt(max_len))

    for num_steps in range(0, max_len):
        if len(h) == 0: break

        h = h.extend(await model.sample(h, num_samples=n, temperature=temperature))
        h = await model.rewrite(h)
        h, done = (h + done).separate_by(not_done)

        h = dc.topk(h, n, name = "h_" + str(num_steps))
        done = dc.topk(done, n, name = "done_" + str(num_steps))

        # stop when done already tracks topk results
        if len(done) == n and dc.max_score(h) < dc.min_score(done):
            break

        yield (h, done)
    
    yield (h, done)

@dc.decoder
async def beam_search(prompt_ids: np.ndarray, n=4, max_len=None, **kwargs):
    n = kwargs.get("num_beams", n)
    max_len = max_len or 2048
    model = dc.model(**kwargs)

    # keep track of active beams and finished sequences

    h = dc.seqs([dc.seq(prompt_ids)] * 1)
    done = dc.seqs()

    # stopping criteria for the beam search
    not_done = dc.logical_and(dc.logical_not(dc.eos), dc.lt(max_len))

    for num_steps in range(0, max_len):
        if len(h) == 0: break

        h = h.extend(await model.topk_continuations(h, k=n))
        h = await model.rewrite(h)
        h, done = (h + done).separate_by(not_done)

        h = dc.topk(h, n, name = "h_" + str(num_steps))
        done = dc.topk(done, n, name = "done_" + str(num_steps))

        # stop when done already tracks topk results
        if len(h) == 0 or (len(done) == n and dc.max_score(h) < dc.min_score(done)):
            break

        yield (h, done)
    
    dc.finish(done)

    yield (h, done)

dc.decoder(beam_search, "beam")

@dc.decoder
async def beam_var(prompt_ids: np.ndarray, n=4, max_len=None, inject_stop=False, prune=None, **kwargs):
    s = Stats("beam_var")

    n = kwargs.get("num_beams", n)
    max_len = max_len or 2048
    model = dc.model(**kwargs)
    alpha = kwargs.get("alpha", 0.7)
    scorer = alpha_length_normalized_det(alpha=alpha)

    # keep track of active beams and finished sequences
    s_initial = dc.seq(prompt_ids)
    s_initial.data("scorer_alpha", alpha, sticky=True)

    h = dc.seqs([s_initial] * 1)
    done = dc.seqs()

    # stopping criteria for the beam search
    not_done = dc.logical_and(dc.logical_not(dc.eos), dc.lt(max_len))

    num_completed = 0

    for num_steps in range(0, max_len):
        if len(h) == 0: break

        with s.timer("det_ext"):
            # generate deterministic parts until a variable is encountered
            det, non_det = h.separate_by(dc.next_is_deterministic)
            while len(det) > 0:
                det = det.extend(await model.argmax(det))
                det = await model.rewrite(det)
                det, non_det = (non_det + det).separate_by(dc.next_is_deterministic)

            h = non_det
            h, done_new = (h + done).separate_by(not_done)
            num_completed += len(done_new) - len(done)
            done = done_new

        h_before_extension = h
        h_extended = h.extend(await model.topk_continuations(h, k=n))

        if inject_stop:
            h_injected = await h_before_extension.aelement_wise(stop_phrase_injection, model, **kwargs)
            h_extended = (h_injected + h_extended)
            h_extended_shape = h_extended.shape
            h_extended = dc.token_unique(h_extended, lambda s, t: dc.is_deterministic(s), flatten=True)
            h_extended = h_extended.reshape(h_extended_shape)

        h = await model.rewrite(h_extended)
        h, done_new = (h + done).separate_by(not_done)
        num_completed += len(done_new) - len(done)
        done = done_new

        if len(done) >= 1 and prune is not None:
            # sscorer = alpha_length_normalized()
            sscorer = sum_scorer()
            ref_score = dc.max_score(done, sscorer) * prune
            det_h, nondet_h = h.separate_by(dc.is_deterministic)
            nondet_h, pruned = nondet_h.separate_by(lambda s: sscorer(s.logprobs, s) > ref_score)
            h = det_h + nondet_h
            print(f"Pruning in step {num_steps} with score {ref_score:.3f}, removed {len(pruned)}")

        # stopping criteria
        if len(h) == 0 or num_completed >= 2*n or (len(done) >= n and dc.max_score(h, scorer=scorer) < dc.min_score(done, scorer=scorer)):
            break

        # additionally pool by variable
        h = h.reshape("head.variable")

        with s.timer("allocation"):
            # keep only topk per pool in h and done
            det_h, nondet_h = h.separate_by(dc.is_deterministic)
            if len(nondet_h) > 0:
                nondet_h = post_vilar_allocation(nondet_h, n, scorer=scorer, num_steps=num_steps)
            h = det_h + nondet_h
            # h = post_vilar_allocation(h, n, scorer=scorer, num_steps=num_steps)

            done = dc.topk(done, n, scorer=scorer, name="done_" + str(num_steps))

        yield (h, done)
    
    done.name("final result")
    yield done
    dc.finish(done)


def post_vilar_allocation(h, k, scorer=None, num_steps=0):
    """
    Restricts h to a set of candidate sequences according to the criteria in Post 
    and Vilar (Fast Lexically Constrained Decoding with Dynamic Beam Allocation for Neural Machine Translation)

    Paper: http://arxiv.org/abs/1804.06609
    
    Choose candidates according to these rules:
    - keep the best top k sequences across all dimensions of h
    - keep all deterministically extended sequences
    - for each sequence of the previous step, keep at least the best continuation
    
    """
    original_shape = h.shape

    # extract deterministic sequences, including stop phrase insertions if active
    det_h, nondet_h = h.separate_by(dc.is_deterministic)
    det_h.name(f"constraint_cont_{num_steps}", nopath=True)

    # Get single best continuation of every sequence
    continuations_by_predecessor = h.reshape(lambda s: s.predecessor.id)
    best_cont_per_predecessor = dc.topk(continuations_by_predecessor, 1, scorer=scorer)
    best_cont_per_predecessor = best_cont_per_predecessor.name(f"best_cont_{num_steps}", nopath=True)
    # best_cont_per_predecessor = dc.topk(best_cont_per_predecessor, k, scorer=scorer).name(f"best_cont_{num_steps}", nopath=True)

    # Get the k best overall sequences
    h_top_k = dc.topk(h.flatten(), k, scorer=scorer)
    h_top_k = h_top_k.reshape(*original_shape).name(f"topk_{num_steps}")

    h_pool = dc.token_unique((h_top_k + det_h + best_cont_per_predecessor).flatten(), lambda s, t: dc.is_deterministic(s))

    # Filter by comparing among sequences satisfying the same number of constraints.
    num_constraints = np.array([s.num_constraints for s in h_pool.unstructured()])
    num_constraints = {det_len: (num_constraints == det_len).sum() for det_len in set(num_constraints.tolist())}


    # print("\nStep Number: ", num_steps)
    beam_size_constrained = 1*k
    bank_sizes = get_bank_sizes_post_vilar(num_constraints, beam_size_constrained)

    pools_by_const_num = h_pool.separate_by_list(lambda s: s.num_constraints)
    h_filtered = DataArray({})
    for num_const, seqs in pools_by_const_num.items():
        seqs_filtered = dc.topk(seqs, bank_sizes[num_const], scorer=scorer).name(f"bank_{num_steps}_{num_const}")
        assert all([s.num_constraints == num_const for s in seqs_filtered.unstructured()])
        h_filtered += seqs_filtered

    h = h_filtered.reshape(*original_shape)

    return h


def get_bank_sizes_post_vilar(num_seq_with_constraints: Dict[int, int],
                   beam_size: int):
    """
    First assign each bank, corresponding to a number of satisfied constraints the same number of slots,
    with the remainder going to the one for the most satisfied constraints. Then, redistribute slots of
    oversized banks downward.
    """

    num_constraints = sorted(list(num_seq_with_constraints.keys()))
    num_banks = len(num_constraints)
    bank_size = beam_size // num_banks
    remainder = beam_size - bank_size * num_banks

    bank_sizes = {det_len: bank_size for det_len in num_constraints}
    bank_sizes[num_constraints[-1]] += remainder

    bank_size_keys = sorted(list(bank_sizes.keys()))

    # reallocate oversized banks
    # print("pre: ", bank_sizes)
    roll_over = 0
    for n_const in bank_size_keys:
        bank_sizes[n_const] += roll_over
        roll_over = 0
        overfill = bank_sizes[n_const] - (num_seq_with_constraints[n_const] if n_const in num_seq_with_constraints else 0)
        if overfill > 0:
            bank_sizes[n_const] -= overfill
            roll_over += overfill

    # print("mid: ", bank_sizes)

    for n_const in reversed(bank_size_keys):
        bank_sizes[n_const] += roll_over
        roll_over = 0
        overfill = bank_sizes[n_const] - (num_seq_with_constraints[n_const] if n_const in num_seq_with_constraints else 0)
        if overfill > 0:
            bank_sizes[n_const] -= overfill
            roll_over += overfill

    # print("post: ", bank_sizes)
    assert sum(bank_sizes.values()) <= beam_size, f"Beam size {beam_size}, total banks: {sum(bank_sizes.values())}"
    return bank_sizes

async def stop_phrase_injection(seqs, model, **kwargs):
    seqs_extended = []

    for s in seqs:
        stop_phrases = s.data("head.stopping_phrases.tokenized")
        if stop_phrases is None or len(stop_phrases) == 0: continue
        if next_is_deterministic(s): continue
        if s.data("injected_stop_phrase"): continue # do not extend in a stop phrase

        for stop_phrase in stop_phrases:
            stop_phrase = np.array(stop_phrase).reshape(-1)
            # consider detseq of stop phrase
            stop_phrase_seq = await model.score(s, stop_phrase)
            # allows to distinguish between stop phrase insertions and normal continuations
            stop_phrase_seq.data("injected_stop_phrase", True)
            seqs_extended.append(stop_phrase_seq)
    return seqs_extended

class shared:
    def __init__(self, v):
        self.v = v
    def set(self, v):
        self.v = v
    def get(self):
        return self.v
    def __repr__(self):
        return str(self)
    def __str__(self):
        return f"shared({self.v})"

class non_rewritten_score(dc.topk_scorer):
    def __init__(self, alpha=0.7):
        self.alpha = alpha

    def __call__(self, scores, s, *args, **kwargs):
        return self.score(scores, self.alpha, s)

    @classmethod
    def score(cls, scores, alpha, s, **kwargs):
        return s.data("eos_score")
        # print("eos_score", s.data("eos_score"))
        # if len(scores) == 0:
        #     return torch.tensor(0.0)
        # else:
        #     return (torch.tensor(1.0) / float(len(scores[:-1])) ** alpha) * scores[:-1].sum()


async def topk_var_continuations(model, seqs: dc.DataArray, active_variable, b, method: str, **kwargs):
    """
    model: model to use for scoring
    seqs: sequences to extend
    active_variable: active variable that should be decoded
    b: number of continuation candidates to generate per sequence
    """
    active = seqs.reshape(lambda s: s.id)
    variable_done = dc.seqs()
    active_variable.set(None)

    def is_active_variable(s):
        v = s.data("head.variable")
        if v is None: return False
        if active_variable.get() is None: 
            active_variable.set(v)
            return True
        return v == active_variable.get()

    temperature = kwargs.get("temperature", 1.0)

    if method == "beam":
        i = 0

        while not is_seq_beams_search_done(active, variable_done, num_beams=b):
            # inner variable decoding (beam_sample with 2*n beams and branching factor n)
            if sample:
                active = active.extend(await model.sample(active, temperature=temperature, num_samples=b))
            else:
                active = active.extend(await model.topk_continuations(active, k=b))
            
            active = await model.rewrite(active)

            active, variable_done = (active + variable_done).separate_by(is_active_variable)
            active = dc.topk(active, b)

            active.name(str(active_variable.get()) + "[" + str(i) + "]_")
            i += 1
            
            yield active
    elif method == "beam_seq":
        i = 0

        regular_scorer = alpha_length_normalized()
        nw_score = non_rewritten_score()
        top_variable_done = dc.seqs()

        while not (len(active) == 0 or (len(top_variable_done) == b and dc.max_score(active, scorer=nw_score) < dc.min_score(top_variable_done, scorer=nw_score))):
            # inner variable decoding (beam_sample with 2*n beams and branching factor n)
            if sample:
                active = active.extend(await model.sample(active, temperature=temperature, num_samples=b))
            else:
                active = active.extend(await model.topk_continuations(active, k=b))
            
            for s in active.flatten().items():
                # s.data("eos_score", s.score
                s.data("eos_score", regular_scorer(s.logprobs))

            active = await model.rewrite(active)

            active, variable_done = (active + variable_done).separate_by(is_active_variable)
            active = dc.topk(active, b)

            top_variable_done, _ = dc.seperate_topk(variable_done, b, scorer=nw_score)

            active.name(str(active_variable.get()) + "[" + str(i) + "]_")
            i += 1
            
            yield active
    elif method == "sample":
        active = active.element_wise(lambda s: s.copy() * b)
        i = 0

        while not is_seq_beams_search_done(active, variable_done, num_beams=b):
            active = active.extend(await model.sample(active, temperature=temperature, num_samples=1))
            
            active = await model.rewrite(active)
            
            active, variable_done = (active + variable_done).separate_by(is_active_variable)
            active = dc.topk(active, b)

            active.name(str(active_variable.get()) + "[" + str(i) + "]_")
            i += 1

            yield active

    # final result should be token unique
    yield dc.token_unique(variable_done)

@dc.decoder
async def bsseq(prompt_ids: np.ndarray, b=2, max_len=384, **kwargs):
    kwargs.pop("n", None)
    async for s in var(prompt_ids, b=b, n=1, max_len=max_len, subdecoder="beam_seq", **kwargs):
        yield s

@dc.decoder
async def var(prompt_ids: np.ndarray, b=2, n=None, max_len=384, subdecoder="sample", **kwargs):
    """
    sample: 
        if True, uses beam_sample variable-local-decoding, otherwise uses beam_search-like decoding
    var_temperature: 
        sampling temperature to use to obtain multiple variable values
    """
    model = dc.model(**kwargs)

    # keep track of active beams and finished sequences
    variable_done = dc.seqs([dc.seq(prompt_ids)])
    active_variable = shared(None)
    done = dc.seqs()
    is_not_done = dc.logical_and(dc.logical_not(dc.eos), dc.lt(max_len))

    n_variable = 0
    # keep same number of beams as branching factor
    if n is None: n = b

    while True:
        # remove sequences that violate constraints
        # variable_done, _ = variable_done.separate_by(dc.is_lmql_valid)
        # if len(variable_done) == 0: break
        
        # generate variable continuation
        async for variable_done in topk_var_continuations(model, variable_done, active_variable, b, subdecoder, **kwargs):
            yield variable_done
        
        # restrict to top b best variable continuations
        variable_done = variable_done.flatten()
        variable_done, done = (done + variable_done).separate_by(is_not_done)
        if len(variable_done) == 0: break

        variable_done = dc.topk(variable_done, b)

        # generate deterministic parts until a variable is encountered
        det, non_det = variable_done.separate_by(dc.next_is_deterministic)
        while len(det) > 0:
            det = det.extend(await model.argmax(det))
            det = await model.rewrite(det)
            det, non_det = (det + non_det).separate_by(dc.next_is_deterministic)
            
            yield (det, non_det)

        # if after deterministic expansion all sequences are done, we are done
        variable_done, done = (det + non_det + done).separate_by(is_not_done)
        
        if len(variable_done) == 0: break

        variable_done = dc.topk(variable_done, n)
        variable_done.name("candidates_" + active_variable.get())

        yield variable_done
        # variable_done = non_det

    done = dc.topk(done, len(done))
    done.name("final result")
    yield done
    
    dc.finish(done)

def is_seq_beams_search_done(active_hypotheses, done_hypotheses, num_beams, scorer=None):
    return len(active_hypotheses) == 0 or (len(done_hypotheses) == num_beams and dc.max_score(active_hypotheses, scorer=scorer) < dc.min_score(done_hypotheses, scorer=scorer))




from functools import reduce
import itertools
from parser.ast import Expression, FunctionCall, Identifier
from lib.state import ProgramState

flatten_unique = lambda xs: list(set(reduce(lambda a, b: a + b, xs)))

def jit(e: Expression, state: ProgramState):
    code, names = compile_and_track(e)
    expr_lambda = eval("lambda " + ",".join(names) + ": " + code, state.globals)
    return lambda state: expr_lambda(*[state.resolve(n) for n in names])

def compile_and_track(e: Expression):
    if type(e) is Identifier:
        return e.name, [e.name]
    elif type(e) is FunctionCall:
        arg_results = [compile_and_track(a) for a in e.args]
        args = ",".join([a[0] for a in arg_results])
        names = flatten_unique([a[1] for a in arg_results])
        return f"{e.function_name}({args})", names


"""
Serves a transformers model as LMQL inference API.
"""

from dataclasses import dataclass, field
from collections import defaultdict

import json
from http.server import BaseHTTPRequestHandler, HTTPServer
from multiprocessing import Queue as MPQueue
from queue import Empty
from queue import Queue
import multiprocessing
from typing import Dict
import requests
import asyncio
import sys
import atexit
import argparse
import time
import os
import subprocess

from transformers import AutoModelForCausalLM, AutoTokenizer, pipeline
import torch

@dataclass
class InferenceServerState:
    model_identifier : str
    tokenizer_descriptor : str
    dtype: str

    queue: Queue
    tokenize_queue: Queue
    all_results_queue : Queue
    
    sample_count: int = 0
    client_results_queues: Dict[str,Queue] = field(default_factory=dict)
    
    exit: bool = False

class TokenizerProcessor:
    def __init__(self, state: InferenceServerState):
        self.model_identifier = state.tokenizer_descriptor
        self.queue = state.tokenize_queue
        self.state = state

    def shutdown(self):
        self.state.exit = True

    def tokenize(self, tokenizer, sample_id, client_id, item):
        text = item["text"]

        if text == "<EOS>":
            input_ids = [tokenizer.eos_token_id]
        elif text == "<BOS>":
            input_ids = [tokenizer.bos_token_id]
        else:
            input_ids = tokenizer(text)["input_ids"]

        self.state.all_results_queue.put({
            "sample_id": sample_id,
            "client_id": client_id,
            "input_ids": input_ids
        })

    def detokenize(self, tokenizer, sample_id, client_id, item):
        input_ids = item["input_ids"]

        text = tokenizer.decode(input_ids)
        self.state.all_results_queue.put({
            "sample_id": sample_id,
            "client_id": client_id,
            "text": text
        })

    def run(self, index):
        # load tokenizer
        tokenizer = AutoTokenizer.from_pretrained(self.model_identifier)
        print("Tokenizer #{} {} ready!".format(index, self.model_identifier))

        while not self.state.exit:
            item = self.queue.get()
            if item is None:
                time.sleep(0.1)
                continue

            sample_id = item["sample_id"]
            client_id = item["client_id"]
            action = item["action"]

            if action == "tokenize":
                self.tokenize(tokenizer, sample_id, client_id, item)
            elif action == "detokenize":
                self.detokenize(tokenizer, sample_id, client_id, item)
            else:
                print("error: unknown TokenizerProcessor action {}".format(action))
        
        print("Tokenizer #{} shut down.".format(index))

    def run_in_parallel(self, n=2):
        atexit.register(self.shutdown)
        
        workers = []

        for i in range(n):
            p = multiprocessing.Process(target=self.run, args=(i,))
            p.start()
            workers.append(p)
        
        return workers

class ModelProcessor:
    def __init__(self, state: InferenceServerState, cuda: bool = False, cache: str = None):
        self.model_identifier = state.model_identifier
        self.queue = state.queue
        self.state = state
        self.cuda = cuda
        
        self.cache = None
        if cache is not None:
            from rocksdict import Rdict
            self.cache = Rdict(cache)
            
        self.request_count = 0
        self.requests_cached = 0
        self.last_report = time.time()
        self.last_request_count = 0
        
        try:
            self.nvidia_logging = subprocess.Popen(["nvidia-smi"], stdout=subprocess.PIPE).wait() == 0
        except:
            self.nvidia_logging = False

    def shutdown(self):
        self.state.exit = True
    
    def __del__(self):
        if self.cache is not None:
            self.cache.close()
        
    def print_stats(self):
        if self.nvidia_logging:
            visible_devices = os.environ.get("CUDA_VISIBLE_DEVICES", None)
            cmds = ["nvidia-smi"]
            if visible_devices is not None:
                cmds.append("-i={}".format(visible_devices))
            cmds += ["--query-gpu=name,memory.used,memory.total,utilization.gpu", "--format=csv,noheader"]
            output = [l.split(", ") for l in subprocess.check_output(cmds).decode("utf-8").split("\n") if l.strip() != ""]
            gpu_usage = ["GPU {} {}, util {}".format(i, row[1] + "/" + row[2], row[3]) for i, row in enumerate(output)]
        else:
            gpu_usage = ["GPU monitoring not available on non-CUDA systems"]
        
        print(" " * 100, end="\r")
        # fancy unicode based terminal spinner
        terminal_spinner_chars = ["⠋", "⠙", "⠹", "⠸", "⠼", "⠴", "⠦", "⠧", "⠇", "⠏"]
        throughput = (self.request_count - self.last_request_count) / (time.time() - self.last_report)
        self.last_report = time.time()
        self.last_request_count = self.request_count
        
        # format throughput to two decimal places
        print("{} {:.2f} calls/s, Requests Served: {}, Queue: {} [{}]".format(
            terminal_spinner_chars[self.request_count % len(terminal_spinner_chars)], 
            throughput,
            self.request_count, 
            self.state.queue.qsize(), 
            ", ".join(gpu_usage)), end="\r")

    def run(self):
        dtype = self.state.dtype
        if dtype == "float16":
            dtype = torch.float16
        else:
            dtype = None
        
        # load model
        if not self.cuda:
            print("Loading {} (CPU)".format(self.model_identifier))
            self.model = AutoModelForCausalLM.from_pretrained(self.model_identifier, torch_dtype=dtype, resume_download=True)
        else:
            print("Loading {} (Multi-GPU)".format(self.model_identifier))
            self.model = AutoModelForCausalLM.from_pretrained(self.model_identifier, torch_dtype=dtype, resume_download=True, device_map="auto")
        self.model.eval()
        
        print("Ready!".format(self.model_identifier))

        while not self.state.exit:
            self.print_stats()
            # wait for self.queue to have an item
            try:
                item = self.queue.get(timeout=1.0)
            except Empty:
                continue
            except KeyboardInterrupt:
                break

            if item is None: 
                time.sleep(0.1)
                continue

            self.request_count += 1
        
            device = "cuda" if self.cuda else "cpu"

            sample_id = item["sample_id"]
            client_id = item["client_id"]
            input_ids = torch.tensor(item["input_ids"], dtype=torch.long).to(device)
            attention_mask = item.get("attention_mask", None)
            
            if attention_mask is None:
                attention_mask = torch.ones_like(input_ids).to(device)
            else:
                attention_mask = torch.tensor(attention_mask, dtype=torch.long).to(device)

            if self.cache is not None:
                key = "IDs:" + str(input_ids.tolist()) + " MASK:" + str(attention_mask.tolist())
                if key in self.cache:
                    self.requests_cached += 1
                    self.state.all_results_queue.put({
                        "client_id": client_id,
                        "sample_id": sample_id,
                        "next_token_logits": self.cache[key]
                    })
                    continue
            
            res = self.model.forward(input_ids=input_ids, attention_mask=attention_mask)
            
            if input_ids.ndimension() == 2:
                next_token_logits = res.logits[:,-1]
            else:
                next_token_logits = res.logits[-1]
            
            if self.cache is not None:
                key = "IDs:" + str(input_ids.tolist()) + " MASK:" + str(attention_mask.tolist())
                self.cache[key] = next_token_logits.tolist()

            self.state.all_results_queue.put({
                "client_id": client_id,
                "sample_id": sample_id,
                "next_token_logits": next_token_logits.detach().tolist()
            })
        
        print("Processor shut down")

    def oom_reloading_run(self):
        while True:
            try:
                self.run()
                return
            except RuntimeError as e:
                if "CUDA out of memory" in str(e):
                    print("Crashed due to OOM, reloading model.")
                    continue
                else:
                    import traceback
                    traceback.print_exc()
                    print("Crashed with", e, "reloading model...")
                    continue

    def run_in_parallel(self):
        atexit.register(self.shutdown)

        p = multiprocessing.Process(target=self.oom_reloading_run)
        p.start()
        return p

class LMQLInferenceAPIHandler(BaseHTTPRequestHandler):
    def __init__(self, *args, **kwargs):
        self._client_id = None
        super().__init__(*args, **kwargs)

    # disable logging
    def log_message(self, format, *args):
        return

    @property
    def state(self) -> InferenceServerState:
        return self.server.state
    
    @property
    def client_id(self) -> str:
        if self._client_id is None:
            self.error_bad_request(msg="client_id not set (please provide the client_id in POST payload or in query.")
            raise Exception("client_id not set")
        return self._client_id

    def error_bad_request(self, msg="Bad request."):
        self.send_response(400)
        self.send_header('Content-type', 'text/plain')
        self.end_headers()
        self.wfile.write(msg.encode("utf-8"))

    def process_some_all_results(self, max=10):
        all_results_queue = self.state.all_results_queue
        i = 0
        
        while not all_results_queue.empty() and i < max:
            result = all_results_queue.get()

            result_client_id = result["client_id"]
            
            if result_client_id not in self.state.client_results_queues:
                self.state.client_results_queues[result_client_id] = Queue()
            self.state.client_results_queues[result_client_id].put(result)

            i += 1

    def do_GET_results(self):
        request_client_id = self.path.split("/")[2]
        self._client_id = f"{self.client_address[0]}-{request_client_id}"

        self.send_response(200)
        self.send_header('Content-type', 'application/json')
        self.end_headers()

        # process some results from the model and group them by client_id
        self.process_some_all_results()

        if self.client_id not in self.state.client_results_queues.keys():
            self.wfile.write(b'[]')
            return

        # return all results for self.client_id currently available
        self.wfile.write("[".encode())
        while not self.state.client_results_queues[self.client_id].empty():
            result = self.state.client_results_queues[self.client_id].get()
            self.wfile.write(json.dumps(result).encode())
            
            # omit last colon
            if self.state.client_results_queues[self.client_id].empty(): break
            else: self.wfile.write(b",")
        self.wfile.write(b"]")

    def do_queue_forward(self, payload, sample_id):
        try: 
            input_ids = payload['input_ids']
            attention_mask = payload.get('attention_mask', None)
            model_identifier = payload['model_identifier']

            if model_identifier != self.state.model_identifier:
                self.error_bad_request("The inference API serves model {} not {}.".format(self.state.model_identifier, model_identifier))
                return

            self._client_id = f"{self.client_address[0]}-{payload['client_id']}"

            assert type(input_ids) == list # and all([type(i) == int for i in input_ids])

            self.state.queue.put({
                'client_id': self.client_id,
                'sample_id': sample_id,
                'input_ids': input_ids,
                'attention_mask': attention_mask
            })

            self.send_response(200)
            self.send_header('Content-type', 'application/json')
            self.end_headers()
            self.wfile.write(json.dumps({'sample_id': sample_id}).encode())
        except Exception as e: 
            self.error_bad_request()

    def do_queue_tokenize(self, payload, sample_id):
        try: 
            text = payload['text']
            self._client_id = f"{self.client_address[0]}-{payload['client_id']}"

            assert type(text) == str

            self.state.tokenize_queue.put({
                'client_id': self.client_id,
                'sample_id': sample_id,
                'action': 'tokenize',
                'text': text
            })

            self.send_response(200)
            self.send_header('Content-type', 'application/json')
            self.end_headers()
            self.wfile.write(json.dumps({'sample_id': sample_id}).encode())
        except Exception as e: 
            self.error_bad_request()
        
    def do_queue_detokenize(self, payload, sample_id):
        try: 
            input_ids = payload['input_ids']
            self._client_id = f"{self.client_address[0]}-{payload['client_id']}"

            assert type(input_ids) == list and all([type(i) == int for i in input_ids])

            self.state.tokenize_queue.put({
                'client_id': self.client_id,
                'sample_id': sample_id,
                'action': 'detokenize',
                'input_ids': input_ids
            })

            self.send_response(200)
            self.send_header('Content-type', 'application/json')
            self.end_headers()
            self.wfile.write(json.dumps({'sample_id': sample_id}).encode())
        except Exception as e: 
            self.error_bad_request()

    def do_POST_queue(self):
        # handle POST to /queue
        payload = self.rfile.read(int(self.headers['Content-Length']))
        payload = json.loads(payload)
        # get client address and port
        sample_id = payload["sample_id"]

        action = payload['action']

        if action == "forward":
            self.do_queue_forward(payload, sample_id)
        elif action == "tokenize":
            self.do_queue_tokenize(payload, sample_id)
        elif action == "detokenize":
            self.do_queue_detokenize(payload, sample_id)
        else:
            self.error_bad_request("Unknown action: {}".format(action))

    def do_POST(self):
        if self.path == "/queue":
            self.do_POST_queue()
            return
        else:
            self.send_error(404)
            return

    def do_GET(self):
        if self.path.startswith("/results"):
            self.do_GET_results()
            return
        else:
            self.send_error(404)
            return

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("model", type=str)
    parser.add_argument("--tokenizer", type=str, default=None)
    parser.add_argument("--port", type=int, default=8080)
    parser.add_argument("--host", type=str, default="localhost")
    parser.add_argument("--cuda", action="store_true", default=False)
    parser.add_argument("--cache", type=str, default=None)
    parser.add_argument("--dtype", type=str, default="none")
    parser.add_argument("--num-tokenizer-processes", type=int, default=2, dest="num_tokenizer_processes")
    
    args = parser.parse_args()
    
    manager = multiprocessing.Manager()
    
    # prepare configuration
    model_descriptor = args.model
    tokenizer_descriptor = args.tokenizer
    if tokenizer_descriptor is None:
        tokenizer_descriptor = model_descriptor
    state = InferenceServerState(model_descriptor, 
                                 tokenizer_descriptor, 
                                 args.dtype, 
                                 queue=manager.Queue(), 
                                 tokenize_queue=manager.Queue(),
                                 all_results_queue=manager.Queue())

    # run model in separate process
    processor = ModelProcessor(state, cuda=args.cuda, cache=args.cache)
    processor.run_in_parallel()

    # run tokenizers in separate process
    tokenizer_processor = TokenizerProcessor(state)
    tokenizer_processor.run_in_parallel(n=args.num_tokenizer_processes)

    # run inference API server in this process
    server_address = (args.host, args.port)
    httpd = HTTPServer(server_address, LMQLInferenceAPIHandler)
    httpd.state = state
    
    try:
        print("Serving LMQL inference API on {}:{}".format(args.host, args.port))
        httpd.serve_forever()
    except KeyboardInterrupt:
        # terminate server
        httpd.shutdown()
        httpd.server_close()
        print("Server stopped")

    # terminate processors
    processor.shutdown()
    tokenizer_processor.shutdown()



from typing import List

from lmql.tests.expr_test_utils import *
from lmql.ops.ops import StopAtOp, NextToken

# enable_show_transformed()

# @lmql_test
# def test_stops_at():
#     expr = LMQLExpr(STOPS_AT(SEQ, "abc here"))
    
#     result = expr.digest("abc here we go abc here")
#     assert tuple(result) == (var(True), fin(False), var(True), var(True), var(True), fin(False))

# @lmql_test
# def test_stops_at2():
#     expr = LMQLExpr(STOPS_AT(SEQ, "abc her"))
    
#     result = expr.digest("abc here we go abc here")
#     assert tuple(result) == (var(True), fin(False), var(True), var(True), var(True), fin(False))

@lmql_test
def test_stops_at1_follow():
    expr = LMQLExpr(STOPS_AT(SEQ, "abc her"))
    
    follow = tuple(expr.follow("abc here we go"))
    assert str(follow) == "(* -> var(True), {eos} -> var(True) and * -> fin(False), * -> var(True), * -> var(True))"

@lmql_test
def test_stops_at2_follow():
    expr = LMQLExpr(STOPS_AT(SEQ, "abc her"))
    
    follow = tuple(expr.follow("abc here we go abc here"))
    assert str(follow) == "(* -> var(True), {eos} -> var(True) and * -> fin(False), * -> var(True), * -> var(True), * -> var(True), {eos} -> var(True) and * -> fin(False))"

run_all_tests(globals())


def get_js_tokenizer(model_identifier):
    import js
    from pyodide.ffi import to_js

    assert "gpt" in model_identifier, "JS tokenizer only supports GPT models."

    class JSBridgedTokenizer:
        """ Custom tokenizer to be used only in a browser environment. This tokenizer only supports GPT tokenization. """
        def __init__(self):
            self.bos_token_id = 50256
            self.eos_token_id = 50256
            self._vocab = None

        @property
        def vocab_size(self):
            return len(self.vocab)

        @property
        def vocab(self):
            if self._vocab is None:
                self._vocab = js.get_vocab().to_py()
            return self._vocab

        def convert_tokens_to_string(self, tokens):
            return js.convert_tokens_to_string_gpt(to_js(tokens))

        def tokenize(self, s):
            unpack = False
            if type(s) is not list:
                s = [s]
                unpack = True
            tokens = [js.tokenize_gpt_toks(se).to_py() for se in s]
            
            if unpack:
                return tokens[0]
            else:
                return tokens
         
        def decode(self, input_ids):
            # set typed array type of input_ids to int
            return str(js.detokenize_gpt(to_js([int(i) for i in input_ids])))

        def __call__(self, s: str):
            unpack = False
            if type(s) is not list:
                s = [s]
                unpack = True
            input_ids = [[int(v) for v in js.tokenize_gpt(se)] for se in s]
            
            if unpack:
                return {"input_ids": input_ids[0]}
            else:
                return {"input_ids": input_ids}
    
    return JSBridgedTokenizer()

global special_token_mappings
special_token_mappings = {}
global reverse_special_token_mappings
reverse_special_token_mappings = {}

class LMQLTokenizer:
    def __init__(self, tokenizer_impl):
        self.tokenizer_impl = tokenizer_impl

        self.detokenizer_cache = {}

    @property
    def vocab_size(self):
        return self.tokenizer_impl.vocab_size

    @property
    def bos_token_id(self):
        return self.tokenizer_impl.bos_token_id
    
    @property
    def eos_token_id(self):
        return self.tokenizer_impl.eos_token_id

    @property
    def vocab(self):
        return self.tokenizer_impl.vocab

    def convert_tokens_to_string(self, tokens):
        return self.tokenizer_impl.convert_tokens_to_string(tokens)

    def tokenize(self, s):
        tokens = []
        for s in self.chunk_out_by_tags(s, tokenize=False):
            if s.startswith("lmql:"):
                tokens.append(s)
            else:
                tokens += self.tokenizer_impl.tokenize(s)
        return tokens
        
    def decode(self, input_ids):
        key = str(input_ids)
        n = len(input_ids)
        if n in self.detokenizer_cache.keys():
            if key in self.detokenizer_cache[n].keys():
                # print("cache hit")
                return self.detokenizer_cache[n][key]
        if n-1 in self.detokenizer_cache.keys():
            key = str(input_ids[:-1])
            if key in self.detokenizer_cache[n-1].keys():
                global reverse_special_token_mappings
                # print("secondary cache hit")
                if input_ids[-1] >= self.tokenizer_impl.vocab_size:
                    extended = self.detokenizer_cache[n-1][key] + "<" + reverse_special_token_mappings[input_ids[-1]] + "/>"
                else:
                    extended = self.detokenizer_cache[n-1][key] + self.tokenizer_impl.decode([input_ids[-1]])
                if not n in self.detokenizer_cache.keys():
                    self.detokenizer_cache[n] = {}
                self.detokenizer_cache[n][str(input_ids)] = extended
                return extended

        s = ""
        for chunk in self.chunk_out_by_special_ids(input_ids):
            if type(chunk) is str:
                s += chunk
            else:
                s += self.tokenizer_impl.decode(chunk)

        if not n in self.detokenizer_cache.keys():
            self.detokenizer_cache[n] = {}
        self.detokenizer_cache[n][key] = s

        return s

    def __call__(self, s: str):
        input_ids = []
        unpack = False
        if type(s) is not list:
            s = [s]
            unpack = True

        for seq in s:
            chunk_input_ids = []
            for chunk in self.chunk_out_by_tags(seq):
                if type(chunk) is int:
                    chunk_input_ids.append(chunk)
                else:
                    chunk_input_ids += self.tokenizer_impl(chunk)["input_ids"]
            input_ids.append(chunk_input_ids)
        
        if unpack:
            return {"input_ids": input_ids[0]}
        else:
            return {"input_ids": input_ids}
    
    def special_token_id(self, identifier):
        global special_token_mappings
        global reverse_special_token_mappings
        
        if identifier not in special_token_mappings:
            if len(special_token_mappings) == 0:
                # offset vocabulary IDs by at least the next decimal power of 10
                offset = 10 ** (len(str(self.vocab_size)))
                special_token_mappings[identifier] = offset
                reverse_special_token_mappings[offset] = identifier
            else:
                next_id = max(special_token_mappings.values()) + 1
                special_token_mappings[identifier] = next_id
                reverse_special_token_mappings[next_id] = identifier
        return special_token_mappings[identifier]
    
    def chunk_out_by_special_ids(self, input_ids, tokenize=True):
        global reverse_special_token_mappings
        c = []
        for i in input_ids:
            if i in reverse_special_token_mappings.keys():
                if len(c) > 0:
                    yield c
                c = []
                yield "<" + reverse_special_token_mappings[i] + "/>"
            else:
                c.append(i)
        yield c
    
    def chunk_out_by_tags(self, s, tokenize=True):
        # filter out all special tokens <lmql:.../>
        import re
        segments = []
        offset = 0
        for m in re.finditer(r"<lmql:(.*?)\/>", s):
            segments.append(s[offset:m.start()])
            if tokenize:
                segments.append(self.special_token_id("lmql:" + m.group(1)))
            else:
                segments.append("lmql:" + m.group(1))
            offset = m.end()
        segments.append(s[offset:])
        return segments

def load_tokenizer(model_identifier):
    import os

    # check environment of USE_JS_TOKENIZER
    if "LMQL_BROWSER" in os.environ:
        return LMQLTokenizer(get_js_tokenizer(model_identifier))

    from transformers import AutoTokenizer
    import torch

    # first try to load pickled tokenizer from cache (faster)
    import pickle
    import pathlib

    cache_dir = pathlib.Path.home() / ".cache" / "lmql"
    cache_dir.mkdir(parents=True, exist_ok=True)
    cache_identifier = model_identifier.replace("/", "-")
    cache_path = cache_dir / f"tokenizer-{cache_identifier}.pkl"

    if cache_path.exists():
        with open(cache_path, "rb") as f:
            return LMQLTokenizer(pickle.load(f))
    else:
        from transformers import AutoTokenizer
        t = AutoTokenizer.from_pretrained(model_identifier)
        with open(cache_path, "wb") as f:
            pickle.dump(t, f)
        return LMQLTokenizer(t)

if __name__ == "__main__":
    import sys

    model_identifier = sys.argv[1]
    t = load_tokenizer(model_identifier)

    to_tokenize = sys.argv[2]

    if to_tokenize.startswith("["):
        import json
        to_tokenize = json.loads(to_tokenize)
        print(str([t.decode(torch.tensor(to_tokenize))])[1:-1])
    else:
        res = t(to_tokenize)
        print(res)
        print(t.convert_ids_to_tokens(res["input_ids"]))
        n = 0
        result = ""
        for t,id in sorted(t.vocab.items(), key=lambda p: p[1]):
            # contains digit
            digits = "0123456789"
            if len(t) < 4 and any(c in digits for c in t):
                print(t,id)
                n += 1
                result += f""""{t}","""
        print(n)
        print(result)


from .dclib_global import *
from .dclib_array import *
from .dclib_model import *
from .dclib_seq import *

class _Decoders: pass
_Decoders.registry = {}

def get_decoder(name):
    return _Decoders.registry[name]

def get_all_decoders():
    return _Decoders.registry.keys()

# decoder decorator which keeps track of all decoders by name
def decoder(fct, name=None):
    if name is None: 
        name = fct.__name__
    _Decoders.registry[name] = fct
    return fct


from lmql.ops.token_set import *
import lmql.ops.token_set as token_set
from transformers import AutoTokenizer

def assert_equal(x, y):
    if x != y:
        raise AssertionError(f"{x} != {y}")

if __name__ == "__main__":
    """
    TODO: tests here need to be fixed (do not use set, but tset)
    """

    VocabularyMatcher.init(AutoTokenizer.from_pretrained("gpt2"))

    # print(tset("abc"))
    # print(tset("abc").mask.sum())

    # print(ntset("abc"))

    # print(tset("abc").union("*"))
    # print(tset("a"))
    # print(tset("abc").union(tset("a", "b", "c")))

    # print(tset("eos"))

    # # print(tset("[^\u0120].*", regex=True).__str__(full=True))
    # print(tset(".*\n.*", regex=True).__str__(full=True))

    # assert intersect(set(["abc"]), set(["cba"])) == "∅"
    
    # p1 = set(["a", "b"])
    # p2 = set(["b", "c"])

    # assert intersect(p1, p2) == set(["b"])

    # p1 = "*"
    # p2 = set(["b", "c"])

    # assert str(sorted(list(intersect(p1, p2)))) == "['b', 'c']"
    # assert str(sorted(list(intersect(p2, p1)))) == "['b', 'c']"

    # assert intersect(set(["abc"]), "∅") == "∅"

    # # f1 = fmap(
    # #     ("eos", True),
    # #     ("*", False)
    # # )
    # # class DummyOp:
    # #     def follow(self, v):
    # #         return "IS_TRUE" if v == True else "IS_FALSE"
    # # fm = follow_apply(f1, DummyOp())
    # # assert fm.value("eos") == "IS_TRUE"
    # # assert fm.value("something else") == "IS_FALSE"

    # # f1 = fmap(
    # #     ("eos", True),
    # #     ("*", False)
    # # )
    # # class DummyOp2:
    # #     def follow(self, v):
    # #         if v == True:
    # #             return fmap(
    # #                 ("eos", "IS_EOS_TRUE"),
    # #                 ("*", "IS_OTHER_TRUE")
    # #             )
    # #         else:
    # #             return fmap(
    # #                 ("eos", "IS_EOS_FALSE"),
    # #                 ("*", "IS_OTHER_FALSE")
    # #             )
    # # fm = follow_apply(f1, DummyOp2())

    # assert tset("b", "c").intersect(tset("a", "b")) == tset("b")
    # assert ntset("b", "c").intersect(tset("a", "b")) == tset("a")
    # assert tset("a", "b").intersect(ntset("b", "c")) == tset("a")
    # assert ntset("b", "c").intersect(ntset("a", "b")) == ntset("a", "b", "c")


    # assert tset("a", "b").intersect("*") == tset("a", "b")
    # assert_equal(tset("a", "b").intersect("∅"), "∅")
    # assert_equal(ntset("a", "b").intersect("*"), ntset("a","b"))
    # assert_equal(ntset("a", "b").intersect("∅"), "∅")

    # assert_equal(tset("a", "b").union("*"), "*")
    # assert_equal(tset("a", "b").union("∅"), tset("a", "b"))
    # assert_equal(ntset("a", "b").union("*"), "*")
    # assert_equal(ntset("a", "b").union("∅"), ntset("a", "b"))
    
    # assert tset("b", "c").union(tset("a", "b")) == tset("a", "b", "c")
    # assert tset("b", "c").union(ntset("a", "b")) == ntset("a")
    # assert ntset("b", "c").union(tset("a", "b")) == ntset("c")
    # assert ntset("b", "c").union(ntset("a", "b")) == ntset("b")


from lmql.tests.expr_test_utils import *

# enable_show_transformed()

@lmql_test
def test_stops_at():
    expr = LMQLExpr(STARTS_WITH(SEQ, ["___ _"]))
    
    result = tuple(expr.digest("abc here we go"))
    assert result == (fin(False), fin(False), fin(False), fin(False))

    result = tuple(expr.digest("___ _ abc"))
    assert result == (var(False), fin(True), fin(True))

    follow = expr.follow("___ abc here we go")
    assert str(next(follow)) == "{ _} -> fin(True) and * \ { _} -> fin(False)"
    assert str(next(follow)) == "* -> fin(False)"
    assert str(next(follow)) == "* -> fin(False)"
    assert str(next(follow)) == "* -> fin(False)"

    follow = expr.follow("___ _ abc here we go")
    # print(tuple(follow))
    assert str(next(follow)) == "{ _} -> fin(True) and * \ { _} -> fin(False)"
    assert str(next(follow)) == "* -> fin(True)"
    assert str(next(follow)) == "* -> fin(True)"
    assert str(next(follow)) == "* -> fin(True)"

run_all_tests(globals())


from typing import Iterable, Tuple
from itertools import product
from lmql.ops.token_set import *

class FollowMap(Iterable):
    def __init__(self):
        self.components = []

    def add(self, component, pattern):
        if type(pattern) is str and pattern != "*":
            pattern = tset(pattern)
        elif type(pattern) is set:
            pattern = tset(*list(pattern))
        self.components.append((pattern, component))

    def __repr__(self):
        return str(self)

    def __str__(self) -> str:
        if len(self.components) == 0:
            return "∅"

        def format_pattern(p):
            if type(p) is set and len(p) == 1:
                return list(p)[0]
            if type(p) is set:
                return sorted(list(p))
            return p

        def format_component(c):
            if type(c) is tuple and len(c) == 2 and type(c[1]) is tuple and len(c[1]) == 1 and c[1][0] in set(["inc", "dec", "fin", "var"]):
                # if c[1][0] == "var":
                #     return c[0]
                return f"{c[1][0]}({c[0]})"
            return c

        return " and ".join(f"{format_pattern(pattern)} -> {format_component(component)}" for pattern, component in self.components)
    
    def __eq__(self, other: object) -> bool:
        assert type(other) is FollowMap, "Can only compare (==) FollowMap instances with other FollowMap instances, not {}".format(type(other))

        return str(other) == str(self)

    def value(self, key):
        for p,c in self:
            if type(p) is set and key in p: 
                return c
            if p == "*": 
                return c
        return None


    def intersect(self, p1):
        """
        TODO: think about intersecting

        (eos) -> None and * -> (How, t)

        with eos

        result should be:

        eos -> None
        
        """

        fm = FollowMap()
        handled = tset()

        for p2, component in self.components:
            assert handled != "*", "Cannot intersect further patterns if '*' has already been handled."
            
            p1 = setminus(p1, handled)
            # p2 = setminus(p2, handled)
            # print("intersect", p1, "and", p2, end="  ")
            intersected_pattern = intersect(p1, p2)
            # print(" gives", intersected_pattern, end=" ")
            intersected_pattern = setminus(intersected_pattern, handled)
            # print("minus", handled, " ", intersected_pattern)

            # empty patterns can be skipped
            if intersected_pattern == "∅": continue

            fm.components.append((intersected_pattern, component))
            
            handled = handled.union(intersected_pattern)
            # handled = union(tset(handled), p2)

        # print("intersect fmap", self, "with", p1, "results in", fm)

        return fm

    def add_all(self, other_map):
        for patter, component in other_map.components:
            self.components.append((patter, component))

    def __iter__(self):
        return ((p,c) for p,c in self.components)

    def hashable(self, value):
        if type(value) is tuple:
            return tuple(self.hashable(t) for t in value)
        elif type(value) is ArgTuple:
            return ArgTuple(self.hashable(t) for t in value)
        elif type(value) is list:
            return tuple(self.hashable(t) for t in value)
        elif type(value) is set:
            return tuple(self.hashable(t) for t in value)
        else:
            return value

    def simplify(self):
        by_value = {}
        
        for p,c in self:
            # make component hashable (i.e. tuples => lists)
            c = self.hashable(c)
            if c in by_value:
                by_value[c] = union(by_value[c], p)
            else:
                by_value[c] = p
        
        # # by_pattern = {}
        # # for c,p in by_value.items():
        # #     if p in by_pattern:
        # #         assert by_pattern[p] == c, "Non-unique mapping for pattern {}: {} and {}".format(p, by_pattern[p], c)
        # #     else:
        # #         by_pattern[p] = [c]

        self.components = [(p,c) for c,p in by_value.items()]

    def product(self, *others):
        if len(others) == 0: return self

        result_map = fmap()
        handled = tset()

        pairings = product(*[list(self)] + [list(o) for o in others])

        for mappings in pairings:
            p_intersected = intersect(*[p for p,v in mappings])

            if setminus(p_intersected, handled) == "∅":
                continue

            if p_intersected == "∅": continue
            result_map.add(ArgTuple(v for p,v in mappings), p_intersected)

            handled = handled.union(p_intersected)
        
        result_map.simplify()

        return result_map

    def map(self, fct):
        mappings = [(pattern, fct(pattern, value)) for pattern, value in self.components]
        return fmap(*mappings)

def zip_fmap(*fmaps):
    """
    Simple zip of multiple fmaps (number of components and 
    patterns must match for all fmaps).
    """

    assert len(fmaps) > 0, "cannot zip_fmap with 0 follow maps."
    
    mappings = []
    for entries in zip(*fmaps):
        patterns = [e[0] for e in entries]
        values = tuple([e[1] for e in entries])
        
        assert all([p == patterns[0] for p in patterns])

        mappings.append((patterns[0], values))
    return fmap(*mappings)

def fmap_product(*maps):
    assert len(maps) != 0, "Cannot construct product of zero follow maps."
    return maps[0].product(*maps[1:])

def final_aware_iterator(follow_map):
    for pattern, item in follow_map:
        if type(item) is ArgTuple:
            yield (pattern, (ArgTuple([v[0] for v in item]), [v[1][0] for v in item]))
        else:
            yield pattern, item

class PredeterminedFinal:
    def __init__(self, value, final):
        self.value = value
        self.final = final
    
    def __str__(self):
        return f"{self.final}({self.value})"

def follow_apply(follow_map, op, op_result, context=None):
    value_follow = FollowMap() 

    for pattern, (value, final) in final_aware_iterator(follow_map):
        if type(value) is ArgTuple:
            result_map = op.follow(*value, context=context, result=op_result)
        else:
            result_map = op.follow(value, context=context, result=op_result)

        if result_map is None:
            result_map = fmap(("*", None)).intersect(pattern)
        elif type(result_map) is FollowMap:
            result_map = result_map.intersect(pattern)
        else:
            result_map = fmap(("*", result_map)).intersect(pattern)

        # include final results per component in the result map
        def determine_final(pattern, result):
            if type(result) is PredeterminedFinal:
                return (result.value, [result.final])
            final_unpacked = final
            if type(final) is tuple:
                final_unpacked = list(final)
            return (result, [op.final(final_unpacked, context=context, operands=value, result=result, pattern=pattern)])

        result_map = result_map.map(determine_final)

        value_follow.add_all(result_map)

    value_follow.simplify()
    
    # print("value_follow", value_follow)

    return value_follow

def all_fmap(mapping):
    if type(mapping) is FollowMap: return mapping

    m = FollowMap()
    m.add(mapping, "*")
    return m

def fmap(*mappings):
    m = FollowMap()
    for pattern, result in mappings:
        m.add(result, pattern)
    return m


import os
import sys

# include ./_ext in path
sys.path.insert(0, os.path.abspath('_ext'))

# Configuration file for the Sphinx documentation builder.

# -- Project information

project = 'LMQL'
copyright = '2023, LMQL Language Team'

release = '0.1'
version = '0.1.0'

# -- General configuration

extensions = [
    'sphinx.ext.duration',
    'sphinx.ext.doctest',
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.intersphinx',
    "myst_parser",
    "lmql_snippets",
    "nbsphinx"
]

source_suffix = {
    '.rst': 'restructuredtext',
    '.txt': 'markdown',
    '.md': 'markdown',
}

pygments_style = 'one-dark'

intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
    'sphinx': ('https://www.sphinx-doc.org/en/master/', None),
}
intersphinx_disabled_domains = ['std']

templates_path = ['_templates']

# -- Options for HTML output

html_theme = 'sphinx_book_theme'

html_favicon = "_static/images/lmql.svg"
html_title = ""
html_theme_options = {
    "home_page_in_toc": True,
}

# -- Options for EPUB output
epub_show_urls = 'footnote'

html_static_path = ['_static']
html_logo = "logo.png"
html_favicon = "lmql.svg"


import os

if "LMQL_BROWSER" in os.environ:
    import lmql.runtime.maiohttp as aiohttp
else:
    import aiohttp


import json
import time
import asyncio

from lmql.runtime.tokenizer import load_tokenizer

class OpenAIStreamError(Exception): pass
class OpenAIRateLimitError(OpenAIStreamError): pass

class Capacity: pass
Capacity.total = 32000 # defines the total capacity available to allocate to different token streams that run in parallel
# a value of 80000 averages around 130k tok/min on davinci with beam_var (lower values will decrease the rate and avoid rate limiting)
Capacity.reserved = 0

stream_semaphore = None

class CapacitySemaphore:
    def __init__(self, capacity):
        self.capacity = capacity

    async def __aenter__(self):
        # wait for self.capacity > capacity
        while True:
            if Capacity.reserved >= Capacity.total:
                await asyncio.sleep(0.5)
            else:
                Capacity.reserved += self.capacity
                break

    async def __aexit__(self, *args):
        Capacity.reserved -= self.capacity

class concurrent:
    def __init__(self, task):
        self.task = task

    def __enter__(self):
        self.task = asyncio.create_task(self.task)
        return self
    
    def __exit__(self, *args):
        self.task.cancel()

async def complete(**kwargs):
    if kwargs["model"].startswith("gpt-3.5-turbo") or "gpt-4" in kwargs["model"]:
        async for r in chat_api(**kwargs): yield r
    else:
        async for r in completion_api(**kwargs): yield r

global tokenizer
tokenizer = None

def tokenize(text):
    global tokenizer
    if tokenizer is None:
        tokenizer = load_tokenizer("gpt2")
    return tokenizer(text)["input_ids"]

def detokenize(input_ids):
    global tokenizer
    if tokenizer is None:
        tokenizer = load_tokenizer("gpt2")
        
    while len(input_ids) > 0 and input_ids[0] == tokenizer.bos_token_id:
        input_ids = input_ids[1:]
    while len(input_ids) > 0 and input_ids[-1] == tokenizer.eos_token_id:
        input_ids = input_ids[:-1]
        
    if len(input_ids) == 0:
        return ""
        
    return tokenizer.decode(input_ids)

def tagged_segments(s):
    import re
    segments = []
    current_tag = None
    offset = 0
    for m in re.finditer(r"<lmql:(.*?)\/>", s):
        if m.start() - offset > 0:
            segments.append({"tag": current_tag, "text": s[offset:m.start()]})
        current_tag = m.group(1)
        offset = m.end()
    segments.append({"tag": current_tag, "text": s[offset:]})
    return segments

async def chat_api(**kwargs):
    from lmql.runtime.openai_secret import openai_secret, openai_org
    global stream_semaphore

    num_prompts = len(kwargs["prompt"])
    max_tokens = kwargs.get("max_tokens", 0)

    assert "logit_bias" not in kwargs.keys(), f"Chat API models do not support advanced constraining of the output, please use no or less complicated constraints."
        

    # transform prompt into chat API format
    prompt_ids = kwargs["prompt"]
    kwargs["prompt"] = [detokenize(p) for p in kwargs["prompt"]]
    
    echo = kwargs.pop("echo")
    
    if echo:
        data = {
            "choices": [
                {
                    "text": p,
                    "index": i,
                    "finish_reason": None,
                    "logprobs": {
                        "text_offset": [0 for t in prompt_tokens],
                        "token_logprobs": [0.0 for t in prompt_tokens],
                        "tokens": [[t] for t in prompt_tokens],
                        "top_logprobs": [{t: 0.0} for t in prompt_tokens]
                    }
                }
             for i,(p,prompt_tokens) in enumerate(zip(kwargs["prompt"], prompt_ids)) ]
        }
        yield data
    
    if max_tokens == 0:
        return
    
    assert len(kwargs["prompt"]) == 1, f"chat API models do not support batched processing"
    
    messages = []
    for s in tagged_segments(kwargs["prompt"][0]):
        role = "user"
        tag = s["tag"]
        if tag == "system":
            role = "system"
        elif tag == "assistant":
            role = "assistant"
        elif tag == "user":
            role = "user"
        elif tag is None:
            role = "user"
        else:
            print(f"warning: {tag} is not a valid tag for the OpenAI chat API. Please use one of :system, :user or :assistant.")
        
        messages.append({
            "role": role, 
            "content": s["text"]
        })
    
    del kwargs["prompt"]
    kwargs["messages"] = messages
    
    del kwargs["logprobs"]

    async with CapacitySemaphore(num_prompts * max_tokens):
        
        current_chunk = ""
        stream_start = time.time()
        
        async with aiohttp.ClientSession() as session:
            async with session.post(
                "https://api.openai.com/v1/chat/completions",
                headers={
                    "Authorization": f"Bearer {openai_secret}",
                    "Content-Type": "application/json",
                },
                json={**kwargs},
            ) as resp:
                last_chunk_time = time.time()
                sum_chunk_times = 0
                n_chunks = 0
                current_chunk_time = 0

                async def chunk_timer():
                    nonlocal last_chunk_time, sum_chunk_times, n_chunks, current_chunk_time
                    while True:
                        await asyncio.sleep(0.5)
                        current_chunk_time = time.time() - last_chunk_time
                        # print("Average chunk time:", sum_chunk_times / n_chunks, "Current chunk time:", current_chunk_time)
                        # print("available capacity", Capacity.total - Capacity.reserved, "reserved capacity", Capacity.reserved, "total capacity", Capacity.total, flush=True)

                        if current_chunk_time > 1.5:
                            print("Token stream took too long to produce next chunk, re-issuing completion request. Average chunk time:", sum_chunk_times / max(1,n_chunks), "Current chunk time:", current_chunk_time, flush=True)
                            resp.close()
                            raise OpenAIStreamError("Token stream took too long to produce next chunk.")

                received_text = ""

                with concurrent(chunk_timer()):
                    async for chunk in resp.content.iter_any():
                        chunk = chunk.decode("utf-8")
                        current_chunk += chunk
                        is_done = current_chunk.strip().endswith("[DONE]")
                        
                        while "data: " in current_chunk:
                            chunks = current_chunk.split("data: ")
                            while len(chunks[0]) == 0:
                                chunks = chunks[1:]
                            if len(chunks) == 1:
                                # last chunk may be incomplete
                                break
                            complete_chunk = chunks[0].strip()
                            current_chunk = "data: ".join(chunks[1:])

                            if len(complete_chunk.strip()) == 0: 
                                continue
                            if complete_chunk == "[DONE]": 
                                return
                            
                            n_chunks += 1
                            sum_chunk_times += time.time() - last_chunk_time
                            last_chunk_time = time.time()
                            
                            data = json.loads(complete_chunk)

                            if "error" in data.keys():
                                message = data["error"]["message"]
                                if "rate limit" in message.lower():
                                    raise OpenAIRateLimitError(message + "local client capacity" + str(Capacity.reserved))
                                else:
                                    raise OpenAIStreamError(message + " (after receiving " + str(n_chunks) + " chunks. Current chunk time: " + str(time.time() - last_chunk_time) + " Average chunk time: " + str(sum_chunk_times / max(1, n_chunks)) + ")", "Stream duration:", time.time() - stream_start)

                            choices = []
                            for c in data["choices"]:
                                delta = c["delta"]
                                # skip non-content annotations for now
                                if not "content" in delta:
                                    if len(delta) == 0: # {} indicates end of stream
                                        choices.append({
                                            "text": "",
                                            "index": c["index"],
                                            "finish_reason": c["finish_reason"],
                                            "logprobs": {
                                                "text_offset": [],
                                                "token_logprobs": [],
                                                "tokens": [],
                                                "top_logprobs": []
                                            }
                                        })
                                    continue
                                text = delta["content"]
                                tokens = tokenize((" " if received_text == "" else "") + text)
                                received_text += text
                                # print([text], [received_text])
                                
                                choices.append({
                                    "text": text,
                                    "index": c["index"],
                                    "finish_reason": c["finish_reason"],
                                    "logprobs": {
                                        "text_offset": [0 for _ in range(len(tokens))],
                                        "token_logprobs": [0.0 for _ in range(len(tokens))],
                                        "tokens": [[t] for t in tokens],
                                        "top_logprobs": [{t: 0.0} for t in tokens]
                                    }
                                })
                            data["choices"] = choices

                            yield data
                        
                        if is_done: break
                    
                resp.close()

                if current_chunk.strip() == "[DONE]":
                    return
                
                try:
                    last_message = json.loads(current_chunk.strip())
                    message = last_message.get("error", {}).get("message", "")
                    if "rate limit" in message.lower():
                        raise OpenAIRateLimitError(message + "local client capacity" + str(Capacity.reserved))
                    else:
                        raise OpenAIStreamError(last_message["error"]["message"] + " (after receiving " + str(n_chunks) + " chunks. Current chunk time: " + str(time.time() - last_chunk_time) + " Average chunk time: " + str(sum_chunk_times / max(1, n_chunks)) + ")", "Stream duration:", time.time() - stream_start)
                        # raise OpenAIStreamError(last_message["error"]["message"])
                except json.decoder.JSONDecodeError:
                    raise OpenAIStreamError("Token stream ended unexpectedly.", current_chunk)
    
async def completion_api(**kwargs):
    from lmql.runtime.openai_secret import openai_secret, openai_org
    global stream_semaphore

    num_prompts = len(kwargs["prompt"])
    max_tokens = kwargs.get("max_tokens", 0)

    async with CapacitySemaphore(num_prompts * max_tokens):
        
        current_chunk = ""
        stream_start = time.time()
        
        async with aiohttp.ClientSession() as session:
            async with session.post(
                "https://api.openai.com/v1/completions",
                headers={
                    "Authorization": f"Bearer {openai_secret}",
                    "Content-Type": "application/json",
                },
                json={**kwargs},
            ) as resp:
                last_chunk_time = time.time()
                sum_chunk_times = 0
                n_chunks = 0
                current_chunk_time = 0

                async def chunk_timer():
                    nonlocal last_chunk_time, sum_chunk_times, n_chunks, current_chunk_time
                    while True:
                        await asyncio.sleep(0.5)
                        current_chunk_time = time.time() - last_chunk_time
                        # print("Average chunk time:", sum_chunk_times / n_chunks, "Current chunk time:", current_chunk_time)
                        # print("available capacity", Capacity.total - Capacity.reserved, "reserved capacity", Capacity.reserved, "total capacity", Capacity.total, flush=True)

                        if current_chunk_time > 1.5:
                            print("Token stream took too long to produce next chunk, re-issuing completion request. Average chunk time:", sum_chunk_times / max(1,n_chunks), "Current chunk time:", current_chunk_time, flush=True)
                            resp.close()
                            raise OpenAIStreamError("Token stream took too long to produce next chunk.")

                with concurrent(chunk_timer()):
                    async for chunk in resp.content.iter_any():
                        chunk = chunk.decode("utf-8")
                        current_chunk += chunk
                        is_done = current_chunk.strip().endswith("[DONE]")
                        
                        while "data: " in current_chunk:
                            chunks = current_chunk.split("data: ")
                            while len(chunks[0]) == 0:
                                chunks = chunks[1:]
                            if len(chunks) == 1:
                                # last chunk may be incomplete
                                break
                            complete_chunk = chunks[0].strip()
                            current_chunk = "data: ".join(chunks[1:])

                            if len(complete_chunk.strip()) == 0: 
                                continue
                            if complete_chunk == "[DONE]": 
                                return
                            
                            n_chunks += 1
                            sum_chunk_times += time.time() - last_chunk_time
                            last_chunk_time = time.time()
                            
                            data = json.loads(complete_chunk)

                            if "error" in data.keys():
                                message = data["error"]["message"]
                                if "rate limit" in message.lower():
                                    raise OpenAIRateLimitError(message + "local client capacity" + str(Capacity.reserved))
                                else:
                                    raise OpenAIStreamError(message + " (after receiving " + str(n_chunks) + " chunks. Current chunk time: " + str(time.time() - last_chunk_time) + " Average chunk time: " + str(sum_chunk_times / max(1, n_chunks)) + ")", "Stream duration:", time.time() - stream_start)

                            yield data
                        
                        if is_done: break
                    
                resp.close()

                if current_chunk.strip() == "[DONE]":
                    return
                
                try:
                    last_message = json.loads(current_chunk.strip())
                    message = last_message.get("error", {}).get("message", "")
                    if "rate limit" in message.lower():
                        raise OpenAIRateLimitError(message + "local client capacity" + str(Capacity.reserved))
                    else:
                        raise OpenAIStreamError(last_message["error"]["message"] + " (after receiving " + str(n_chunks) + " chunks. Current chunk time: " + str(time.time() - last_chunk_time) + " Average chunk time: " + str(sum_chunk_times / max(1, n_chunks)) + ")", "Stream duration:", time.time() - stream_start)
                        # raise OpenAIStreamError(last_message["error"]["message"])
                except json.decoder.JSONDecodeError:
                    raise OpenAIStreamError("Token stream ended unexpectedly.", current_chunk)

async def main():
    kwargs = {
        "model": "text-davinci-003",
        "prompt": "Say this is a test",
        "max_tokens": 7,
        "temperature": 0,
        "stream": True
    }

    async for chunk in complete(**kwargs):
        print(chunk)

if __name__ == "__main__":
    import asyncio
    asyncio.run(main())


from lmql.tests.expr_test_utils import *

# enable_show_transformed()

@lmql_test
def test_var():
    expr = LMQLExpr(SEQ)
    result = expr.digest("How are you?".split(" "))

    assert next(result) == inc(("How",))

@lmql_test
def test_len():
    expr = LMQLExpr(len(SEQ))
    result = expr.digest("How are you?".split(" "))

    assert next(result) == inc(1)
    assert next(result) == inc(2)
    assert next(result) == inc(3)

@lmql_test
def test_lt():
    expr = LMQLExpr(len(SEQ) < 2)
    result = expr.digest("How are you?".split(" "))

    assert next(result) == var(True) # [How]
    assert next(result) == fin(False) # [How, are]
    assert next(result) == fin(False) # [How, are, you?]
    
    result = expr.digest("How are you?".split(" "))
    assert next(result) != fin(True) # [How]

@lmql_test
def test_gt():
    expr = LMQLExpr(len(SEQ))
    result = expr.digest("How are you?".split(" "))

    assert next(result) == var(False) # [How]
    assert next(result) == var(False) # [How, are]
    assert next(result) == fin(True) # [How, are, you?]

@lmql_test
def test_gt():
    expr = LMQLExpr(SELECT(SEQ, 1))
    result = expr.digest("How are you?".split(" "))

    assert next(result) == var(None)
    assert next(result) == fin("are")

@lmql_test
def test_len_constant():
    expr = LMQLExpr(len(SEQ) < var(3))
    result = expr.digest("How are you?".split(" "))

    assert next(result) == var(True)
    assert next(result) == var(True)
    assert next(result) == var(False)

@lmql_test
def test_len_gt():
    expr = LMQLExpr(len(SEQ) > 2)
    result = expr.digest("How are you?".split(" "))

    assert next(result) == var(False)
    assert next(result) == var(False)
    assert next(result) == fin(True)
    
@lmql_test
def test_gt_constant():
    expr = LMQLExpr(fin(4) > fin(3))
    result = tuple(expr.digest("How are you?".split(" ")))

    assert result == (fin(True), fin(True), fin(True))

@lmql_test
def test_lower_upper():
    expr = LMQLExpr(len(SEQ) > 2 and len(SEQ) < 5)
    result = tuple(expr.digest("one two three four five six seven".split(" ")))

    assert result == (var(False), var(False), var(True), var(True), fin(False), fin(False), fin(False))

@lmql_test
def test_lower_upper_empty():
    expr = LMQLExpr(len(SEQ) > 5 and len(SEQ) < 2)
    result = tuple(expr.digest("one two three four five six seven".split(" ")))

    assert result == (var(False), fin(False), fin(False), fin(False), fin(False), fin(False), fin(False))

@lmql_test
def test_lower_upper_hole():
    expr = LMQLExpr(len(SEQ) < 2 or len(SEQ) > 4)
    result = tuple(expr.digest("one two three four five six seven".split(" ")))

    # print(result) 
    assert result == (var(True), var(False), var(False), var(False), fin(True), fin(True), fin(True))

run_all_tests(globals())


def gsm8k_samples(): return """Q: Every hour Joanne has to collect the coins out of the fountain inside the mall. During the first hour, she collected 15 coins. For the next two hours, she collected 35 coins from the fountain. In the fourth hour, she collected 50 coins from the fountain but she gave 15 of them to her coworker so she could buy a soda. How many coins did she have after the fourth hour?
A: Let's think step by step.
15 coins collected in hour one
35 coins collected in hour two
35 coins collected in hour three
50 coins collected in hour four
Before giving her coworker some coins there were 15+35+35+50=<< 15+35+35+50 = 135 >>135 coins
The number of coins after given 15 to her coworker is 135-15=<< 135-15 = 120 >>120
So the answer is 120

Q: Jerry’s two daughters play softball on different teams. They each have 8 games this season. Each team practices 4 hours for every game they play. If each game lasts for 2 hours, how many hours will Jerry spend at the field watching his daughters play and practice altogether?
A: Let's think step by step.
Jerry will spend 8 games x 2 hours per game = << 8*2 = 16 >>16 hours watching one daughter play her games.
He will spend 16 x 2 = << 16*2 = 32 >>32 hours watching both daughters play their games.
He will spend 8 games x 4 hours of practice = << 8*4 = 32 >>32 hours watching one daughter practice.
He will spend 32 x 2 = << 32*2 = 64 >>64 hours watching both daughters practice.
He will spend a total of 32 hours watching games + 64 hours watching practice = << 32+64 = 96 >>96 hours.
So the answer is 96
"""



import os
import termcolor

class DebuggerOutputWriter:
    def add_interpreter_head_state(self, variable, head, prompt, where, trace, is_valid, is_final, mask, num_tokens, program_variables): pass
    def add_compiler_output(self, code): pass

class PrintingDebuggerOutputWriter:
    def __init__(self, clear=False):
        self.clear = clear
        self.print_output = True

    def add_decoder_state(*args, **kwargs): 
        pass

    def add_interpreter_head_state(self, variable, head, prompt, where, trace, is_valid, is_final, mask, num_tokens, program_variables):
        if head == 0:
            if self.clear:
                os.system("clear")
            if self.print_output:
                print(f"{prompt}\n\n valid={is_valid}, final={is_final}")
    def add_compiler_output(self, code): pass
    
class StreamingOutputWriter:
    def __init__(self, variable=None):
        self.variable = variable
        
        self.last_value = None
    
    def add_decoder_state(*args, **kwargs): 
        pass

    def add_interpreter_head_state(self, variable, head, prompt, where, trace, is_valid, is_final, mask, num_tokens, program_variables):
        if head == 0:
            if self.variable is not None:
                vars = self.variable
                if type(vars) is not list:
                    vars = [vars]
                
                value = "\n".join(program_variables.variable_values.get(v, "").strip() for v in vars)
                    
                    # os.system("clear")
                if self.last_value is None:
                    self.last_value = value
                    print(value, end="", flush=True)
                else:
                    print(value[len(self.last_value):], end="", flush=True)
                    self.last_value = value
                return
            
            os.system("clear")
            print(f"{prompt}\n", end="\r")
            
    def add_compiler_output(self, code): pass

silent = DebuggerOutputWriter()
stream = StreamingOutputWriter
printing = PrintingDebuggerOutputWriter(clear=True)


import sys
sys.path.append("../")
import lmql
import asyncio
import time

from lmql.runtime.output_writer import PrintingDebuggerOutputWriter

class TestOutputWriter(PrintingDebuggerOutputWriter):
    def __init__(self):
        super().__init__()
        self.clear = False
        self.print_output = False

async def test_argmax_a_joke():
    query = """
argmax
 "A list of good dad jokes. A indicates the punchline \\n"
 "Q: How does a penguin build its house? \\n"
 "A: Igloos it together. END \\n"
 "Q: Which knight invented King Arthur's Round Table? \\n"
 "A: Sir Cumference. END \\n"
 "Q:[JOKE] \\n"
 "A:[PUNCHLINE] \\n"
FROM
   'openai/text-ada-001'
"""
    await lmql.run(query, output_writer=TestOutputWriter())


async def test_argmax_a_joke_with_constraints():
    query = """
argmax
 "A list of good dad jokes. A indicates the punchline \\n"
 "Q: How does a penguin build its house? \\n"
 "A: Igloos it together. END \\n"
 "Q: Which knight invented King Arthur's Round Table? \\n"
 "A: Sir Cumference. END \\n"
 "Q:[JOKE] \\n"
 "A:[PUNCHLINE] \\n"
FROM
   'openai/text-ada-001'
WHERE
   len(JOKE) < 120 and STOPS_AT(JOKE, "?") and 
   STOPS_AT(PUNCHLINE, "END") and len(PUNCHLINE) > 12
"""
    await lmql.run(query, output_writer=TestOutputWriter())

async def test_argmax_a_joke_with_constraints2():
    query = """
argmax
 "A list of good dad jokes. A indicates the punchline \\n"
 "Q: How does a penguin build its house? \\n"
 "A: Igloos it together. END \\n"
 "Q: Which knight invented King Arthur's Round Table? \\n"
 "A: Sir Cumference. END \\n"
 "Q:[JOKE] \\n"
 "A:[PUNCHLINE] another[ENDING] \\n"
FROM
 'openai/text-ada-001'
WHERE
 len(JOKE) < 120 and STOPS_AT(JOKE, "?") and 
 STOPS_AT(PUNCHLINE, "END") and 
 PUNCHLINE in [" A man has", "B"]
"""
    await lmql.run(query, output_writer=TestOutputWriter())


async def test_beam_var():
    query = """
beam_var
   "A list of good dad jokes. A indicates the punchline \\n"
   "Q: How does a penguin build its house? \\n"
   "A: Igloos it together. END \\n"
   "Q: Which knight invented King Arthur's Round Table? \\n"
   "A: Sir Cumference. END \\n"
   "Q:[JOKE] \\n"
   "A:[PUNCHLINE] another[ENDING] \\n"
FROM
   'openai/text-ada-001'
WHERE
   len(JOKE) < 120 and STOPS_AT(JOKE, "?") and 
   STOPS_AT(PUNCHLINE, "END") and 
   PUNCHLINE in [" A man has", "B"]
"""
    await lmql.run(query, output_writer=TestOutputWriter())


async def test_beam_var_hf():
    query = """
beam_var
   "A list of good dad jokes. A indicates the punchline \\n"
   "Q: How does a penguin build its house? \\n"
   "A: Igloos it together. END \\n"
   "Q: Which knight invented King Arthur's Round Table? \\n"
   "A: Sir Cumference. END \\n"
   "Q:[JOKE] \\n"
   "A:[PUNCHLINE] another[ENDING] \\n"
FROM
   'gpt2-medium'
WHERE
   len(JOKE) < 120 and STOPS_AT(JOKE, "?") and 
   STOPS_AT(PUNCHLINE, "END") and 
   PUNCHLINE in [" A man has", "B"]
"""
    await lmql.run(query, output_writer=TestOutputWriter())


async def main():
    # collect all method starting with test_ and run them
    for name, method in globals().items():
        if name.startswith("test_"):
            start = time.time()
            title = f"Running {name}"
            print(title.ljust(120, "."), end="", flush=True)
            await method()
            duration = time.time() - start
            print("(%.2fs)" % duration, " [OK]", flush=True)

asyncio.run(main())


from lmql.language.qstrings import qstring_to_stmts, TemplateVariable
from lmql.language.fragment_parser import LMQLQuery
import ast

class LMQLValidationError(Exception): ...

class QueryStringDistributionVarValidator(ast.NodeVisitor):
    """
    Transformes string expressions on statement level to model queries.
    """
    
    def __init__(self, query, variable_name):
        self.query = query
        self.distribution_variable = variable_name
        self.distribution_variable_used = False
    
    def validate(self):
        for p in self.query.prompt:
            self.visit(p)

    def visit_Expr(self, expr):
        if type(expr.value) is ast.Constant:
            self.check_Constant(expr.value)
        else:
            self.generic_visit(expr)
        return expr

    def check_Constant(self, constant):
        if type(constant.value) is not str: return constant
        qstring = constant.value
        # TODO: handle escaping more completely and gracefully
        qstring = qstring.replace("\n", "\\\\n")
        
        # get 
        for qstmt in qstring_to_stmts(qstring):
            if self.distribution_variable_used:
                raise LMQLValidationError("The distribution variable can only be used at the very end of the prompt.")
            if type(qstmt) is TemplateVariable:
                if qstmt.name == self.distribution_variable:
                    self.distribution_variable_used = True

class LMQLValidator:
    def validate(self, query: LMQLQuery):
        if query.distribution is not None:
            self.check_distribution_var_valid_position(query, query.distribution.variable_name)

    def check_distribution_var_valid_position(self, q: LMQLQuery, variable_name: str):
        QueryStringDistributionVarValidator(q, variable_name).validate()




from lib.errors import error
from parser.ast import FunctionCall, Identifier, Variable
from parser.ast import PromptInput, PromptVariable
from parser.ast import Query

class Scope:
    def __init__(self):
        self.names = {}

    def add(self, name, meaning):
        self.names[name] = meaning

    def has(self, name):
        return name in self.names

    def resolve(self, name):
        return self.names[name]

def scope(query):
    assert query.scope is None, "Cannot call scope() more than once on the same Query element."
    query.scope = Scope()

    # prompt first
    for l in query.prompt_clause:
        if type(l) is PromptInput:
            scope_expr(l.input, query)
        elif type(l) is PromptVariable:
            if query.scope.has(l.name):
                error("redeclaration of output variable {}.".format(l.name), l)

            var = Variable(l.name, "str")
            l.meaning = var
            query.scope.add(l.name, var)

    return query

def scope_expr(e, query):
    if type(e) is str:
        # nothing to scope in string prompts
        return
    elif type(e) is Identifier:
        if not query.scope.has(e.name):
            # automatically introduce input variables (user input fields)
            query.input_variables += [Variable(e.name, "str")]
            e.meaning = query.input_variables[-1]
            # error("failed to resolve variable {}".format(l.name), l)
        else:
            e.meaning = query.scope.resolve(e.name)
    elif type(e) is FunctionCall:
        if not query.scope.has(e.function_name):
            # automatically introduce input variables (user input fields)
            query.input_variables += [Variable(e.function_name, "function")]
            e.meaning = query.input_variables[-1]
            
            for a in e.args: scope_expr(a, query)
            # error("failed to resolve variable {}".format(l.name), l)
    else:
        error("{} expression is not supported as part of prompt".format(type(e)), e)


import time

class StatsTimer:
    def __init__(self, stats, tag):
        self.stats = stats
        self.tag = tag
        self.start = time.time()

    def __enter__(self):
        return self
    
    def __exit__(self, exc_type, exc_value, traceback):
        self.stats.count(self.tag, time.time() - self.start)

class Stats:
    stats = {}

    @staticmethod
    def print_all():
        for name, stats in Stats.stats.items():
            print(f"{name}:")
            print("\n".join(["\t" + line for line in str(stats).split("\n")]))
    
    def __new__(cls, name):
        if name not in cls.stats:
            cls.stats[name] = super().__new__(cls)
        return cls.stats[name]

    def __init__(self, name):
        self.times = {}

    def timer(self, tag):
        return StatsTimer(self, tag)

    def count(self, tag, time):
        if tag not in self.times:
            self.times[tag] = []
        self.times[tag].append(time)
    
    def __str__(self):
        return "\n".join([f"{tag}: {len(times)} calls, {sum(times)}s total, {sum(times)/len(times)}s avg" for tag, times in self.times.items()])

